# coding=utf-8
# This program is in the public domain
# Author: Paul Kienzle
r"""
Experimental probe.

The experimental probe describes the incoming beam for the experiment.
Scattering properties of the sample are dependent on the type and
energy of the radiation.

See :ref:`data-guide` for details.

"""

from __future__ import with_statement, division, print_function

# TOF stitching introduces a lot of complexity.
# Theta offset:
#   Q1 = 4 pi sin(T1 + offset)/L1
#   Q2 = 4 pi sin(T2 + offset)/L2
#   at offset=0,
#      Q1=Q2
#   at offset!=0,
#      Q1' ~ Q1 + 4pi offset/L1
#      Q2' ~ Q2 + 4pi offset/L2
#      => Q1' != Q2'
# Thick layers:
#   Since a given Q, dQ has multiple T, dT, L, dL, oversampling is going
#   to be very complicated.
# Energy dependent SLD:
#   Just because two points are at the same Q does not mean they have
#   the same theory function when scattering length density is
#   energy dependent
#
# Not stitching has its own issues.
# Calculation speed:
#   overlapping points are recalculated
#   profiles are recalculated
#
# Unstitched seems like the better bet.

import os
import json
import warnings
from dataclasses import dataclass
from enum import Enum

import numpy as np
from numpy import sqrt, pi, inf, sign, log
import numpy.random
import numpy.fft
from typing import NamedTuple, Optional, Any, Sequence, Union, TYPE_CHECKING

from bumps.util import field, field_desc, Optional, Any, Union, Dict, Callable, Literal, Tuple, List, Literal

from periodictable import nsf, xsf
from bumps.parameter import Parameter, to_dict, Constant
from bumps.plotutil import coordinated_colors, auto_shift
from bumps.data import parse_multi, strip_quotes
from bumps.util import USE_PYDANTIC

if USE_PYDANTIC or TYPE_CHECKING:
    from bumps.util import NDArray

from . import fresnel
from .material import Vacuum
from .resolution import QL2T, QT2L, TL2Q, dQdL2dT, dQdT2dLoL, dTdL2dQ
from .resolution import sigma2FWHM, FWHM2sigma, dQ_broadening
from .stitch import stitch
from .reflectivity import convolve, BASE_GUIDE_ANGLE
from .util import asbytes

PROBE_KW = (
    "T",
    "dT",
    "L",
    "dL",
    "data",
    "name",
    "filename",
    "intensity",
    "background",
    "back_absorption",
    "sample_broadening",
    "theta_offset",
    "back_reflectivity",
    "data",
)


def make_probe(**kw):
    """
    Return a reflectometry measurement object of the given resolution.
    """
    radiation = kw.pop("radiation")
    kw = dict((k, v) for k, v in kw.items() if k in PROBE_KW)
    if radiation == "neutron":
        return NeutronProbe(**kw)
    else:
        return XrayProbe(**kw)


class BaseProbe:
    intensity: Parameter
    background: Parameter
    back_absorption: Parameter
    Q: "NDArray"
    R: "NDArray"
    name: Optional[str] = None
    dR: Optional[Union[Sequence, "NDArray"]] = None
    dQ: Optional[Union[Sequence, "NDArray"]] = None
    back_reflectivity: bool = False
    resolution: Literal["normal", "uniform"] = "uniform"

    _Ro: Optional[Union[Sequence, "NDArray"]] = field(init=False)

    view = "log"
    plot_shift = 0
    residuals_shift = 0
    show_resolution = True

    @property
    def calc_Q(self):
        """define in derived classes"""
        raise NotImplemented

    def log10_to_linear(self):
        """
        Convert data from log to linear.

        Older reflectometry reduction code stored reflectivity in log base 10
        format.  Call probe.log10_to_linear() after loading this data to
        convert it to linear for subsequent display and fitting.
        """
        if self.R is not None:
            self.R = 10**self.R
            if self.dR is not None:
                self.dR = self.R * self.dR * log(10)

    def resynth_data(self):
        """
        Generate new data according to the model R' ~ N(R, dR).

        The resynthesis step is a precursor to refitting the data, as is
        required for certain types of monte carlo error analysis.  The
        first time it is run it will save the original R into Ro.  If you
        reset R in the probe you will also need to reset Ro so that it
        is used for subsequent resynth analysis.
        """
        if not hasattr(self, "_Ro"):
            self._Ro = self.R
        self.R = self._Ro + numpy.random.randn(*self._Ro.shape) * self.dR

    def restore_data(self):
        """
        Restore the original data after resynth.
        """
        self.R = self._Ro
        del self._Ro

    # CRUFT: Ro doesn't need to be part of the public interface.
    @property
    def Ro(self):
        warnings.warn("Use probe.R instead of probe.Ro.", DeprecationWarning)
        return getattr(self, "_Ro", self.R)

    def simulate_data(self, theory, noise=2.0):
        r"""
        Set the data for the probe to R + eps with eps ~ normal(dR^2).

        *theory* is (Q, R),

        If the percent *noise* is provided, set dR to R*noise/100 before
        simulating.  *noise* defaults to 2% if no dR is present.

        Note that measured data estimates uncertainty from the number of
        counts.  This means that points above the true value will have
        larger uncertainty than points below the true value.  This bias
        is not captured in the simulated data.
        """
        # Minimum value for dR after noise is added.
        # TODO: does this need to be a parameter?
        noise_floor = 1e-11

        # Set the theory function.
        R = np.array(theory[1], "d")  # Force copy
        assert R.shape == self.Q.shape
        self.R = R

        # Make sure scalar noise is positive.  This check is here to so that
        # old interfaces will fail properly.
        if np.isscalar(noise) and noise <= 0.0:
            raise ValueError("Noise level must be positive")

        # If dR is missing then default noise to 2% so that dR will be set.
        if self.dR is None and noise is None:
            noise = 2.0

        # Set dR if noise was given or otherwise defaulted.
        if noise is not None:
            self.dR = 0.01 * np.asarray(noise) * self.R
            self.dR[self.dR < noise_floor] = noise_floor

        # Add noise according to dR.
        self.R += numpy.random.randn(*self.R.shape) * self.dR

    def Q_c(self, substrate=None, surface=None):
        Srho, Sirho = (0, 0) if substrate is None else substrate.sld(self)[:2]
        Vrho, Virho = (0, 0) if surface is None else surface.sld(self)[:2]
        drho = Srho - Vrho if not self.back_reflectivity else Vrho - Srho
        Q_c = sign(drho) * sqrt(16 * pi * abs(drho) * 1e-6)
        return Q_c

    def resolution_guard(self):
        r"""
        Make sure each measured $Q$ point has at least 5 calculated $Q$
        points contributing to it in the range $[-3\Delta Q, 3\Delta Q]$.

        *Not Implemented*
        """
        raise NotImplementedError
        # TODO: implement resolution guard.

    def _apply_resolution(self, Qin, Rin, interpolation):
        """
        Apply the instrument resolution function
        """
        Q, dQ = _interpolate_Q(self.Q, self.dQ, interpolation)
        if np.iscomplex(Rin).any():
            R_real = convolve(Qin, Rin.real, Q, dQ, resolution=self.resolution)
            R_imag = convolve(Qin, Rin.imag, Q, dQ, resolution=self.resolution)
            R = R_real + 1j * R_imag
        else:
            R = convolve(Qin, Rin, Q, dQ, resolution=self.resolution)
        return Q, R

    def apply_beam(self, calc_Q, calc_R, resolution=True, interpolation=0):
        r"""
        Apply factors such as beam intensity, background, backabsorption,
        resolution to the data.

        *resolution* is True if the resolution function should be applied
        to the reflectivity.

        *interpolation* is the number of Q points to show between the
        nominal Q points of the probe. Use this to draw a smooth theory
        line between the data points. The resolution dQ is interpolated
        between the resolution of the surrounding Q points.

        If an amplitude signal is provided, $r$ will be scaled by
        $\surd I + i \surd B / |r|$, which when squared will equal
        $I |r|^2 + B$. The resolution function will be applied directly
        to the amplitude. Unlike intensity and background, the resulting
        $|G \ast r|^2 \ne G \ast |r|^2$ for convolution operator $\ast$,
        but it should be close.
        """
        # Note: in-place vector operations are not notably faster.

        # Handle absorption through the substrate, which occurs when Q<0
        # (condition)*C is C when condition is True or 0 when False,
        # (condition)*(C-1)+1 is C when condition is True or 1 when False.
        back = (calc_Q < 0) * (self.back_absorption.value - 1) + 1
        calc_R = calc_R * back

        # For back reflectivity, reverse the sign of Q after computing
        if self.back_reflectivity:
            calc_Q = -calc_Q
        if calc_Q[-1] < calc_Q[0]:
            calc_Q, calc_R = [v[::-1] for v in (calc_Q, calc_R)]
        if resolution:
            Q, R = self._apply_resolution(calc_Q, calc_R, interpolation)
        else:
            # Given that the target Q points should be in the set of
            # calculated Q values, interp will give us the
            # values of Q at the appropriate R, even if there are
            # guard values, or otherwise some mixture of calculated
            # Q values.  The cost of doing so is going to be n log n
            # in the size of Q, which is a bit pricey, but let's see
            # if it is a problem before optimizing.
            Q, dQ = _interpolate_Q(self.Q, self.dQ, interpolation)
            Q, R = self.Q, np.interp(Q, calc_Q, calc_R)
        if np.iscomplex(R).any():
            # When R is an amplitude you can scale R by sqrt(A) to reproduce
            # the effect of scaling the intensity in the reflectivity. To
            # reproduce the effect of adding a background you can fiddle the
            # phase of r as well using:
            #      s = (sqrt(A) + i sqrt(B)/|r|) r
            # then
            #      |s|^2 = |sqrt(A) + i sqrt(B)/|r||^2 |r|^2
            #            = (A + B/|r|^2) |r|^2
            #            = A |r|^2 + B
            # Note that this cannot work for negative background since
            # |s|^2 >= 0 always, whereas negative background could push the
            # reflectivity below zero.
            R = np.sqrt(self.intensity.value) * R
            if self.background.value > 0:
                R += 1j * np.sqrt(self.background.value) * R / abs(R)
        else:
            R = self.intensity.value * R + self.background.value
        # return calc_Q, calc_R
        return Q, R

    def fresnel(self, substrate=None, surface=None):
        """
        Returns a Fresnel reflectivity calculator given the surface and
        and substrate.  The calculated reflectivity includes The Fresnel
        reflectivity for the probe reflecting from a block of material with
        the given substrate.

        Returns F = R(probe.Q), where R is magnitude squared reflectivity.
        """
        # Doesn't use ProbeCache, but this routine is not time critical
        Srho, Sirho = (0, 0) if substrate is None else substrate.sld(self)[:2]
        Vrho, Virho = (0, 0) if surface is None else surface.sld(self)[:2]
        if self.back_reflectivity:
            Srho, Vrho = Vrho, Srho
            Sirho, Virho = Virho, Sirho
        if Srho == Vrho:
            Srho = Vrho + 1
        # I = np.ones_like(self.Q)
        I = 1
        calculator = fresnel.Fresnel(rho=Srho * I, irho=Sirho * I, Vrho=Vrho * I, Virho=Virho * I)
        return calculator

    def save(self, filename, theory, substrate=None, surface=None):
        """
        Save the data and theory to a file.
        """
        fresnel_calculator = self.fresnel(substrate, surface)
        Q, FQ = self.apply_beam(self.calc_Q, fresnel_calculator(self.calc_Q))
        Q, R = theory
        if len(Q) != len(self.Q):
            # Saving interpolated data
            A = np.array((Q, R, np.interp(Q, self.Q, FQ)))
            header = "# %17s %20s %20s\n" % ("Q (1/A)", "theory", "fresnel")
        elif getattr(self, "R", None) is not None:
            A = np.array((self.Q, self.dQ, self.R, self.dR, R, FQ))
            header = "# %17s %20s %20s %20s %20s %20s\n" % ("Q (1/A)", "dQ (1/A)", "R", "dR", "theory", "fresnel")
        else:
            A = np.array((self.Q, self.dQ, R, FQ))
            header = "# %17s %20s %20s %20s\n" % ("Q (1/A)", "dQ (1/A)", "theory", "fresnel")

        header = ("# intensity: %.15g\n# background: %.15g\n" % (self.intensity.value, self.background.value)) + header

        with open(filename, "wb") as fid:
            # print("saving", A)
            fid.write(asbytes(header))
            np.savetxt(fid, A.T, fmt="%20.15g")

    def plot(self, view=None, **kwargs):
        """
        Plot theory against data.

        Need substrate/surface for Fresnel-normalized reflectivity
        """

        view = view if view is not None else self.view

        if view == "linear":
            self.plot_linear(**kwargs)
        elif view == "log":
            self.plot_log(**kwargs)
        elif view == "fresnel":
            self.plot_fresnel(**kwargs)
        elif view == "logfresnel":
            self.plot_logfresnel(**kwargs)
        elif view == "q4":
            self.plot_Q4(**kwargs)
        elif view == "resolution":
            self.plot_resolution(**kwargs)
        elif view.startswith("resid"):
            self.plot_residuals(**kwargs)
        elif view == "fft":
            self.plot_fft(**kwargs)
        elif view == "SA":  # SA does not plot since it does not exist
            pass
        else:
            raise TypeError("incorrect reflectivity view '%s'" % view)

    def plot_resolution(self, suffix="", label=None, **kwargs):
        import matplotlib.pyplot as plt

        plt.plot(self.Q, self.dQ, label=self.label(prefix=label, suffix=suffix))
        plt.xlabel(r"Q ($\AA^{-1}$)")
        plt.ylabel(r"Q resolution ($1-\sigma \AA^{-1}$)")
        plt.title("Measurement resolution")

    def plot_linear(self, **kwargs):
        """
        Plot the data associated with probe.
        """
        import matplotlib.pyplot as plt

        self._plot_pair(ylabel="Reflectivity", **kwargs)
        plt.yscale("linear")

    def plot_log(self, **kwargs):
        """
        Plot the data associated with probe.
        """
        import matplotlib.pyplot as plt

        self._plot_pair(ylabel="Reflectivity", **kwargs)
        plt.yscale("log")

    def plot_logfresnel(self, *args, **kw):
        """
        Plot the log Fresnel-normalized reflectivity associated with the probe.
        """
        import matplotlib.pyplot as plt

        self.plot_fresnel(*args, **kw)
        plt.yscale("log")

    def plot_fresnel(self, substrate=None, surface=None, **kwargs):
        r"""
        Plot the Fresnel-normalized reflectivity associated with the probe.

        Note that the Fresnel reflectivity has the intensity and background
        applied before normalizing so that hydrogenated samples display more
        cleanly.  The formula to reproduce the graph is:

        .. math::

             R' = R / (F(Q) I + B)
             \Delta R' = \Delta R / (F(Q) I + B)

        where $I$ is the intensity and $B$ is the background.
        """
        if substrate is None and surface is None:
            raise TypeError("Fresnel-normalized reflectivity needs substrate or surface")
        F = self.fresnel(substrate=substrate, surface=surface)

        # print("substrate", substrate, "surface", surface)
        def scale(Q, dQ, R, dR, interpolation=0):
            Q, fresnel = self.apply_beam(self.calc_Q, F(self.calc_Q), interpolation=interpolation)
            return Q, dQ, R / fresnel, dR / fresnel

        if substrate is None:
            name = "air:%s" % surface.name
        elif surface is None or isinstance(surface, Vacuum):
            name = substrate.name
        else:
            name = "%s:%s" % (substrate.name, surface.name)
        self._plot_pair(scale=scale, ylabel="R/(R(%s)" % name, **kwargs)

    def plot_Q4(self, **kwargs):
        r"""
        Plot the Q**4 reflectivity associated with the probe.

        Note that Q**4 reflectivity has the intensity and background applied
        so that hydrogenated samples display more cleanly.  The formula
        to reproduce the graph is:

        .. math::

             R' = R / ( (100*Q)^{-4} I + B)
             \Delta R' = \Delta R / ( (100*Q)^{-4} I + B )

        where $B$ is the background.
        """

        def scale(Q, dQ, R, dR, interpolation=0):
            # Q4 = np.maximum(1e-8*Q**-4, self.background.value)
            Q4 = 1e-8 * Q**-4 * self.intensity.value + self.background.value
            return Q, dQ, R / Q4, dR / Q4

        # Q4[Q4==0] = 1
        self._plot_pair(scale=scale, ylabel="R (100 Q)^4", **kwargs)

    def _plot_pair(
        self,
        theory=None,
        scale=lambda Q, dQ, R, dR, interpolation=0: (Q, dQ, R, dR),
        ylabel="",
        suffix="",
        label=None,
        plot_shift=None,
        **kwargs,
    ):
        import matplotlib.pyplot as plt

        c = coordinated_colors()
        plot_shift = plot_shift if plot_shift is not None else Probe.plot_shift
        trans = auto_shift(plot_shift)
        if hasattr(self, "R") and self.R is not None:
            Q, dQ, R, dR = scale(self.Q, self.dQ, self.R, self.dR)
            if not self.show_resolution:
                dQ = None
            plt.errorbar(
                Q,
                R,
                yerr=dR,
                xerr=dQ,
                capsize=0,
                fmt=".",
                color=c["light"],
                transform=trans,
                label=self.label(prefix=label, gloss="data", suffix=suffix),
            )
        if theory is not None:
            # TODO: completely restructure interpolation handling
            # Interpolation is used to show the theory curve between the
            # data points.  The _calc_Q points used to predict theory at
            # the measured data are used for the interpolated Q points, with
            # the resolution function centered on each interpolated value.
            # The result is that when interpolation != 0, there are more
            # theory points than data points, and we will need to accomodate
            # this when computing normalization curves for Fresnel and Q^4
            # reflectivity.
            # Issues with current implementation:
            # * If the resolution is too tight, there won't be sufficient
            #   support from _calc_Q to compute theory at Q interpolated.
            # * dQ for the interpolated points uses linear interpolation
            #   of dQ between neighbours.  If measurements with tight and
            #   loose resolution are interleaved the result will look very
            #   strange.
            # * There are too many assumptions about interpolation shared
            #   between Experiment and Probe objects.  In particular, the
            #   Fresnel object needs to be computed at the same interpolated
            #   points as the theory function.
            # * If there are large gaps in the data the interpolation will
            #   not fill them in correctly.  Perhaps we should set _Q_plot
            #   and _calc_Q_plot independently from the data?
            # * We sometimes show the theory curve without resolution
            #   applied; this has not been tested with interpolation
            interpolation = kwargs.get("interpolation", 0)
            Q, R = theory
            Q, dQ, R, dR = scale(Q, 0, R, 0, interpolation=interpolation)
            plt.plot(
                Q,
                R,
                "-",
                color=c["dark"],
                transform=trans,
                label=self.label(prefix=label, gloss="theory", suffix=suffix),
            )
            # from numpy.fft import fft
            # x, y = Q[1::2], abs(fft(R)[:(len(R)-1)//2])
            # y = y * (R.max()/y.max())
            # plt.plot(x, y, '-')
        plt.xlabel("Q (inv Angstroms)")
        plt.ylabel(ylabel)
        # plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
        h = plt.legend(fancybox=True, numpoints=1)
        h.get_frame().set_alpha(0.5)

    def plot_residuals(self, theory=None, suffix="", label=None, plot_shift=None, **kwargs):
        import matplotlib.pyplot as plt

        plot_shift = plot_shift if plot_shift is not None else Probe.residuals_shift
        trans = auto_shift(plot_shift)
        if theory is not None and self.R is not None:
            c = coordinated_colors()
            Q, R = theory
            # In case theory curve is evaluated at more/different points...
            R = np.interp(self.Q, Q, R)
            residual = (R - self.R) / self.dR
            plt.plot(
                self.Q, residual, ".", color=c["light"], transform=trans, label=self.label(prefix=label, suffix=suffix)
            )
        plt.axhline(1, color="black", ls="--", lw=1)
        plt.axhline(0, color="black", lw=1)
        plt.axhline(-1, color="black", ls="--", lw=1)
        plt.xlabel("Q (inv A)")
        plt.ylabel("(theory-data)/error")
        plt.legend(numpoints=1)

    def plot_fft(self, theory=None, suffix="", label=None, substrate=None, surface=None, **kwargs):
        """
        FFT analysis of reflectivity signal.
        """
        raise NotImplementedError
        import matplotlib.pyplot as plt

        c = coordinated_colors()
        if substrate is None and surface is None:
            raise TypeError("FFT reflectivity needs substrate or surface")
        F = self.fresnel(substrate=substrate, surface=surface)
        # Qc = sqrt(16*pi*substrate)
        Qc = 0
        Qmax = max(self.Q)
        T = np.linspace(Qc, Qmax, len(self.Q))
        z = np.linspace(0, 2 * pi / Qmax, len(self.Q) // 2)
        if hasattr(self, "R"):
            signal = np.interp(T, self.Q, self.R / F(self.Q))
            A = abs(numpy.fft.fft(signal - np.average(signal)))
            A = A[: len(A) // 2]
            plt.plot(z, A, ".-", color=c["light"], label=self.label(prefix=label, gloss="data", suffix=suffix))
        if theory is not None:
            Q, R = theory
            signal = np.interp(T, Q, R / F(Q))
            A = abs(numpy.fft.fft(signal - np.average(signal)))
            A = A[: len(A) // 2]
            plt.plot(z, A, "-", color=c["dark"], label=self.label(prefix=label, gloss="theory", suffix=suffix))
        plt.xlabel("w (A)")
        if substrate is None:
            name = "air:%s" % surface.name
        elif surface is None or isinstance(surface, Vacuum):
            name = substrate.name
        else:
            name = "%s:%s" % (substrate.name, surface.name)
        plt.ylabel("|FFT(R/R(%s))|" % name)

    def label(self, prefix=None, gloss="", suffix=""):
        if not prefix:
            prefix = self.name
        if prefix:
            return " ".join((prefix + suffix, gloss)) if gloss else prefix
        else:
            return suffix + " " + gloss if gloss else None


@dataclass(init=False)
class Probe(BaseProbe):
    r"""
    Defines the incident beam used to study the material.

    For calculation purposes, probe needs to return the values $Q_\text{calc}$
    at which the model is evaluated.  This is normally going to be the measured
    points only, but for some systems, such as those with very thick layers,
    oversampling is needed to avoid aliasing effects.

    A measurement point consists of incident angle, angular resolution,
    incident wavelength, FWHM wavelength resolution, reflectivity and
    uncertainty in reflectivity.

    A probe is a set of points, defined by vectors for point attribute.  For
    convenience, the attribute can be initialized with a scalar if it is
    constant throughout the measurement, but will be set to a vector in
    the probe.  The attributes are initialized as follows:

        *T* : float or [float] | degrees
            Incident angle
        *dT* : float or [float] | degrees
            FWHM angular divergence
        *L* : float or [float] | |Ang|
            Incident wavelength
        *dL* : float or [float] | |Ang|
            FWHM wavelength dispersion
        *data* : ([float], [float])
            R, dR reflectivity measurement and uncertainty
        *dQ* : [float] or None | |1/Ang|
            1-\$sigma$ Q resolution when it cannot be computed directly
            from angular divergence and wavelength dispersion.
        *resolution* : 'normal' or 'uniform'
            Distribution function for Q resolution.

    Measurement properties:

        *intensity* : float or Parameter
           Beam intensity
        *background* : float or Parameter
           Constant background
        *back_absorption* : float or Parameter
           Absorption through the substrate relative to beam intensity.
           A value of 1.0 means complete transmission; a value of 0.0
           means complete absorption.
        *theta_offset* : float or Parameter
           Offset of the sample from perfect alignment
        *sample_broadening* : float or Parameter
           Additional FWHM angular divergence from sample curvature.
           Scale 1-$\sigma$ rms by $2 \surd(2 \ln 2) \approx 2.35$ to convert
           to FWHM.
        *back_reflectivity* : True or False
           True if the beam enters through the substrate

    Measurement properties are fittable parameters.  *theta_offset* in
    particular should be set using *probe.theta_offset.dev(dT)*, with *dT*
    equal to the FWHM uncertainty in the peak position for the rocking curve,
    as measured in radians. Changes to *theta_offset* will then be penalized
    in the cost function for the fit as if it were another measurement.  Use
    :meth:`alignment_uncertainty` to compute dT from the shape of the
    rocking curve.

    Sample broadening adjusts the existing Q resolution rather than
    recalculating it. This allows it the resolution to describe more
    complicated effects than a simple gaussian distribution of wavelength
    and angle will allow. The calculation uses the mean wavelength, angle
    and angular divergence. See :func:`resolution.dQ_broadening` for details.

    *intensity* and *back_absorption* are generally not needed --- scaling
    the reflected signal by an appropriate intensity measurement will correct
    for both of these during reduction.  *background* may be needed,
    particularly for samples with significant hydrogen content due to its
    large isotropic incoherent scattering cross section.

    """

    # Fields:

    intensity: Parameter
    background: Parameter
    back_absorption: Parameter
    theta_offset: Parameter
    sample_broadening: Parameter
    name: Optional[str] = None
    filename: Optional[str] = None
    back_reflectivity: bool = False
    R: Optional[Any] = None
    dR: Optional[Any] = 0
    T: "NDArray" = field_desc("List of theta values (incident angle)")
    dT: Optional[Any] = 0
    L: "NDArray" = field_desc("List of lambda values (wavelength, in Angstroms)")
    dL: Optional[Any] = 0
    dQo: Optional[Union[Sequence, "NDArray"]] = None
    resolution: Literal["normal", "uniform"] = "uniform"
    oversampling: Optional[int] = None
    oversampling_seed: int = 1
    radiation: Literal["neutron", "xray"] = "xray"

    polarized = False
    Aguide = BASE_GUIDE_ANGLE  # default guide field for unpolarized measurements
    view = "log"
    plot_shift = 0
    residuals_shift = 0
    show_resolution = True

    @classmethod
    def from_dict(cls, **kw):
        R = kw.pop("R", None)
        dR = kw.pop("dR", None)
        if R is not None and dR is not None:
            kw["data"] = (R, dR)
        radiation = kw.pop("radiation")
        if radiation == "neutron":
            return NeutronProbe(**kw)
        else:
            return XrayProbe(**kw)

    def __init__(
        self,
        T=None,
        dT=0,
        L=None,
        dL=0,
        data=None,
        intensity=1,
        background=0,
        back_absorption=1,
        theta_offset=0,
        sample_broadening=0,
        back_reflectivity=False,
        name: Optional[str] = None,
        filename=None,
        dQo=None,
        resolution: Literal["normal", "uniform"] = "normal",
        oversampling=None,
        oversampling_seed=1,
    ):
        if T is None or L is None:
            raise TypeError("T and L required")
        if sample_broadening is None:
            sample_broadening = 0
        if theta_offset is None:
            theta_offset = 0
        if not name and filename:
            name = os.path.splitext(os.path.basename(filename))[0]
        qualifier = " " + name if name is not None else ""
        self.intensity = Parameter.default(intensity, name="intensity" + qualifier)
        self.background = Parameter.default(background, name="background" + qualifier, limits=(0.0, None))
        self.back_absorption = Parameter.default(back_absorption, name="back_absorption" + qualifier, limits=(0.0, 1.0))
        self.theta_offset = Parameter.default(theta_offset, name="theta_offset" + qualifier)
        self.sample_broadening = Parameter.default(sample_broadening, name="sample_broadening" + qualifier)
        self.back_reflectivity = back_reflectivity
        if data is not None:
            R, dR = data
            R = np.array(R)
            dR = np.array(dR)
        else:
            R, dR = None, None

        self._set_TLR(T, dT, L, dL, R, dR, dQo)
        self.name = name
        self.filename = filename
        self.resolution = resolution
        if oversampling is not None:
            self.oversample(oversampling, oversampling_seed)

    def _set_TLR(self, T, dT, L, dL, R, dR, dQ):
        # if L is None:
        #    L = xsf.xray_wavelength(E)
        #    dL = L * dE/E
        # else:
        #    E = xsf.xray_energy(L)
        #    dE = E * dL/L

        T = np.array(T, np.float64)
        L = np.array(L, np.float64)
        Q = TL2Q(T=T, L=L)
        if dQ is not None:
            dQ = np.asarray(dQ)
        else:
            dQ = dTdL2dQ(T=T, dT=dT, L=L, dL=dL)

        # Make sure that we are dealing with vectors
        T, dT, L, dL = [np.ones_like(Q) * v for v in (T, dT, L, dL)]

        # remove nan
        nan_indices = set()
        for column in [T, dT, L, dL, R, dR, Q, dQ]:
            if column is not None:
                indices = np.argwhere(np.isnan(column)).flatten()
                nan_indices.update(indices)

        nan_indices = list(nan_indices)
        T, dT, L, dL, R, dR, Q, dQ = (
            np.delete(c, nan_indices) if c is not None else None for c in [T, dT, L, dL, R, dR, Q, dQ]
        )

        # Probe stores sorted values for convenience of resolution calculator
        idx = np.argsort(Q)
        self.T, self.dT = T[idx], dT[idx]
        self.L, self.dL = L[idx], dL[idx]
        self.Qo, self.dQo = Q[idx], dQ[idx]
        if R is not None:
            R = R[idx]
        if dR is not None:
            dR = dR[idx]
        self.R = R
        self.dR = dR

        # By default the calculated points are the measured points.  Use
        # oversample() for a more accurate resolution calculations.
        self._set_calc(self.T, self.L)

    @staticmethod
    def alignment_uncertainty(w, I, d=0):
        r"""
        Compute alignment uncertainty.

        **Parameters:**

        *w* : float | degrees
            Rocking curve full width at half max.
        *I* : float | counts
            Rocking curve integrated intensity.
        *d* = 0: float | degrees
            Motor step size

        **Returns:**

        *dtheta* : float | degrees
            uncertainty in alignment angle
        """
        return sqrt(w**2 / I + d**2 / 12.0)

    def write_data(self, filename, columns=("Q", "R", "dR"), header=None):
        """
        Save the data to a file.

        *header* is a string with trailing \\n containing the file header.
        *columns* is a list of column names from Q, dQ, R, dR, L, dL, T, dT.

        The default is to write Q, R, dR data.
        """
        if header is None:
            header = "# %s\n" % " ".join(columns)
        with open(filename, "wb") as fid:
            fid.write(asbytes(header))
            data = np.vstack([getattr(self, c) for c in columns])
            np.savetxt(fid, data.T)

    def _set_calc(self, T, L):
        Q = TL2Q(T=T, L=L)

        idx = np.argsort(Q)
        self.calc_T = T[idx]
        self.calc_L = L[idx]
        self.calc_Qo = Q[idx]

        # Only keep the scattering factors that you need
        self.unique_L = np.unique(self.calc_L)
        self._L_idx = np.searchsorted(self.unique_L, L)

    @property
    def Q(self):
        if self.theta_offset.value != 0:
            Q = TL2Q(T=self.T + self.theta_offset.value, L=self.L)
            # TODO: this may break the Q order on measurements with varying L
        else:
            Q = self.Qo
        return Q

    @Q.setter
    def Q(self, Q):
        # If we explicity set Q, then forget what we know about T and L.
        # This will cause theta offset != 0 to fail.
        if hasattr(self, "T"):
            del self.T, self.L
        self.Qo = Q

    @property
    def dQ(self):
        if self.sample_broadening.value == 0:
            dQ = self.dQo
        else:
            dQ = dQ_broadening(dQ=self.dQo, L=self.L, T=self.T, dT=self.dT, width=self.sample_broadening.value)
        return dQ

    @dQ.setter
    def dQ(self, dQ):
        self.dQo = dQ

    @property
    def calc_Q(self):
        if self.theta_offset.value != 0:
            Q = TL2Q(T=self.calc_T + self.theta_offset.value, L=self.calc_L)
            # TODO: this may break the Q order on measurements with varying L
        else:
            Q = self.calc_Qo
        return Q if not self.back_reflectivity else -Q

    def parameters(self):
        return {
            "intensity": self.intensity,
            "background": self.background,
            "back_absorption": self.back_absorption,
            "theta_offset": self.theta_offset,
            "sample_broadening": self.sample_broadening,
        }

    def to_dict(self):
        """Return a dictionary representation of the parameters"""
        return to_dict(
            {
                "type": type(self).__name__,
                "name": self.name,
                "filename": self.filename,
                "intensity": self.intensity,
                "background": self.background,
                "back_absorption": self.back_absorption,
                "theta_offset": self.theta_offset,
                "sample_broadening": self.sample_broadening,
            }
        )

    def scattering_factors(self, material, density):
        """
        Returns the scattering factors associated with the material given
        the range of wavelengths/energies used in the probe.
        """
        raise NotImplementedError("need radiation type in <%s> to compute sld for %s" % (self.filename, material))

    def subsample(self, dQ):
        """
        Select points at most every dQ.

        Use this to speed up computation early in the fitting process.

        This changes the data object, and is not reversible.

        The current algorithm is not picking the "best" Q value, just the
        nearest, so if you have nearby Q points with different quality
        statistics (as happens in overlapped regions from spallation
        source measurements at different angles), then it may choose
        badly.  Simple solutions based on the smallest relative error dR/R
        will be biased toward peaks, and smallest absolute error dR will
        be biased toward valleys.
        """
        # Assumes self contains sorted Qo and associated T, L
        # Note: calc_Qo is already sorted
        Q = np.arange(self.Qo[0], self.Qo[-1], dQ)
        idx = np.unique(np.searchsorted(self.Qo, Q))
        # print len(idx), len(self.Qo)

        self.T, self.dT = self.T[idx], self.dT[idx]
        self.L, self.dL = self.L[idx], self.dL[idx]
        self.Qo, self.dQo = self.Qo[idx], self.dQo[idx]
        if self.R is not None:
            self.R = self.R[idx]
        if self.dR is not None:
            self.dR = self.dR[idx]
        self._set_calc(self.T, self.L)

    def critical_edge(self, substrate=None, surface=None, n=51, delta=0.25):
        r"""
        Oversample points near the critical edge.

        The critical edge is defined by the difference in scattering
        potential for the *substrate* and *surface* materials, or the
        reverse if *back_reflectivity* is true.

        *n* is the number of $Q$ points to compute near the critical edge.

        *delta* is the relative uncertainty in the material density,
        which defines the range of values which are calculated.

        Note: :meth:`critical_edge` will remove the extra Q calculation
        points introduced by :meth:`oversample`.

        The $n$ points $Q_i$ are evenly distributed around the critical
        edge in $Q_c \pm \delta Q_c$ by varying angle $\theta$ for a
        fixed wavelength $< \lambda >$, the average of all wavelengths
        in the probe.

        Specifically:

        .. math::

            Q_c^2 &= 16 \pi (\rho - \rho_\text{incident}) \\
            Q_i &= Q_c - \delta_i Q_c (i - (n-1)/2)
                \qquad \text{for} \; i \in 0 \ldots n-1 \\
            \lambda_i &= < \lambda > \\
            \theta_i &= \sin^{-1}(Q_i \lambda_i / 4 \pi)

        If $Q_c$ is imaginary, then $-|Q_c|$ is used instead, so this
        routine can be used for reflectivity signals which scan from
        back reflectivity to front reflectivity.  For completeness,
        the angle $\theta = 0$ is added as well.
        """
        Q_c = self.Q_c(substrate, surface)
        Q = np.linspace(Q_c * (1 - delta), Q_c * (1 + delta), n)
        L = np.average(self.L)
        T = QL2T(Q=Q, L=L)
        T = np.hstack((self.T, T, 0))
        L = np.hstack((self.L, [L] * (n + 1)))
        # print Q
        self._set_calc(T, L)

    def oversample(self, n=20, seed=1):
        """
        Generate an over-sampling of Q to avoid aliasing effects.

        Oversampling is needed for thick layers, in which the underlying
        reflectivity oscillates so rapidly in Q that a single measurement
        has contributions from multiple Kissig fringes.

        Sampling will be done using a pseudo-random generator so that
        accidental structure in the function does not contribute to the
        aliasing.  The generator will usually be initialized with a fixed
        *seed* so that the point selection will not change from run to run,
        but a *seed* of None will choose a different set of points each time
        oversample is called.

        The value *n* is the number of points that should contribute to
        each Q value when computing the resolution.   These will be
        distributed about the nominal measurement value, but varying in
        both angle and energy according to the resolution function.  This
        will yield more points near the measurement and fewer farther away.
        The measurement point itself will not be used to avoid accidental
        bias from uniform Q steps.  Depending on the problem, a value of
        *n* between 20 and 100 should lead to stable values for the convolved
        reflectivity.

        Note: :meth:`oversample` will remove the extra Q calculation
        points introduced by :meth:`critical_edge`.
        """

        rng = numpy.random.RandomState(seed=seed)
        T = rng.normal(self.T[:, None], self.dT[:, None], size=(len(self.dT), n - 1))
        L = rng.normal(self.L[:, None], self.dL[:, None], size=(len(self.dL), n - 1))
        T = np.hstack((self.T, T.flatten()))
        L = np.hstack((self.L, L.flatten()))
        self._set_calc(T, L)
        self.oversampling = n
        self.oversampling_seed = seed


class XrayProbe(Probe):
    """
    X-Ray probe.

    By providing a scattering factor calculator for X-ray scattering, model
    components can be defined by mass density and chemical composition.
    """

    radiation = "xray"

    # TODO: remove density from interface since it is always 1.0
    # Density is handled as a scale factor applied to the returned sld
    # in material.ProbeCache. This allows us to fit the density without
    # looking up the sld each time.
    def scattering_factors(self, material, density):
        # doc string is inherited from parent (see below)
        # TODO: support wavelength dependent systems
        rho, irho = xsf.xray_sld(material, wavelength=self.unique_L[0], density=density)
        return rho, irho, 0

    scattering_factors.__doc__ = Probe.scattering_factors.__doc__


class NeutronProbe(Probe):
    """
    Neutron probe.

    By providing a scattering factor calculator for X-ray scattering, model
    components can be defined by mass density and chemical composition.
    """

    radiation = "neutron"

    # TODO: remove density from interface since it is always 1.0
    # Density is handled as a scale factor applied to the returned sld
    # in material.ProbeCache. This allows us to fit the density without
    # looking up the sld each time.
    def scattering_factors(self, material, density):
        # doc string is inherited from parent (see below)
        # TODO: support wavelength dependent systems
        rho, irho, rho_incoh = nsf.neutron_sld(material, wavelength=self.unique_L[0], density=density)
        # print(f"{material}@{density} L={self.unique_L[0]} rho={rho}")
        return rho, irho, rho_incoh

    scattering_factors.__doc__ = Probe.scattering_factors.__doc__


class ProbeSet(Probe):
    def __init__(self, probes, name=None):
        self.probes = list(probes)
        self.R = np.hstack([p.R for p in self.probes])
        self.dR = np.hstack([p.dR for p in self.probes])
        self.name = name if name is not None else self.probes[0].name

        back_refls = [f.back_reflectivity for f in self.probes]
        if all(back_refls) or not any(back_refls):
            self.back_reflectivity = back_refls[0]
        else:
            raise ValueError("Cannot mix front and back reflectivity measurements")

    def parameters(self):
        return [p.parameters() for p in self.probes]

    parameters.__doc__ = Probe.parameters.__doc__

    def to_dict(self):
        """Return a dictionary representation of the parameters"""
        return to_dict(
            {
                "type": type(self).__name__,
                "name": self.name,
                "probes": self.probes,
            }
        )

    def resynth_data(self):
        for p in self.probes:
            p.resynth_data()
        self.R = np.hstack([p.R for p in self.probes])

    resynth_data.__doc__ = Probe.resynth_data.__doc__

    def restore_data(self):
        for p in self.probes:
            p.restore_data()
        self.R = np.hstack([p.R for p in self.probes])

    restore_data.__doc__ = Probe.restore_data.__doc__

    def simulate_data(self, theory, noise=2.0):
        """
        Simulate data, allowing for noise to be a dR array for each Q point.
        """
        Q, R = theory
        dR = np.asarray(noise)
        offset = 0
        for p in self.probes:
            n = len(p.Q)
            if len(self.dR.shape) > 0:
                noise = dR[offset : offset + n]
            p.simulate_data(theory=(Q[offset : offset + n], R[offset : offset + n]), noise=noise)
            offset += n

    simulate_data.__doc__ = Probe.simulate_data.__doc__

    @property
    def Q(self):
        return np.hstack([p.Q for p in self.probes])

    @property
    def calc_Q(self):
        return np.unique(np.hstack([p.calc_Q for p in self.probes]))

    @property
    def dQ(self):
        return np.hstack([p.dQ for p in self.probes])

    @property
    def unique_L(self):
        return np.unique(np.hstack([p.unique_L for p in self.probes]))

    def oversample(self, **kw):
        for p in self.probes:
            p.oversample(**kw)

    oversample.__doc__ = Probe.oversample.__doc__

    def scattering_factors(self, material, density):
        # TODO: support wavelength dependent systems
        return self.probes[0].scattering_factors(material, density)
        # result = [p.scattering_factors(material, density) for p in self.probes]
        # return [np.hstack(v) for v in zip(*result)]

    scattering_factors.__doc__ = Probe.scattering_factors.__doc__

    def apply_beam(self, calc_Q, calc_R, interpolation=0, **kw):
        result = [p.apply_beam(calc_Q, calc_R, **kw) for p in self.probes]
        Q, R = [np.hstack(v) for v in zip(*result)]
        return Q, R

    def fresnel(self, *args, **kw):
        return self.probes[0].fresnel(*args, **kw)

    fresnel.__doc__ = Probe.fresnel.__doc__

    def save(self, filename, theory, substrate=None, surface=None):
        for i, (p, th) in enumerate(self.parts(theory=theory)):
            p.save(filename + str(i + 1), th, substrate=substrate, surface=surface)

    save.__doc__ = Probe.save.__doc__

    def plot(self, theory=None, **kw):
        import matplotlib.pyplot as plt

        for p, th in self.parts(theory):
            p.plot(theory=th, **kw)

    plot.__doc__ = Probe.plot.__doc__

    def plot_resolution(self, **kw):
        for p in self.probes:
            p.plot_resolution(**kw)

    plot_resolution.__doc__ = Probe.plot_resolution.__doc__

    def plot_linear(self, theory=None, **kw):
        for p, th in self.parts(theory):
            p.plot_linear(theory=th, **kw)

    plot_linear.__doc__ = Probe.plot_linear.__doc__

    def plot_log(self, theory=None, **kw):
        for p, th in self.parts(theory):
            p.plot_log(theory=th, **kw)

    plot_log.__doc__ = Probe.plot_log.__doc__

    def plot_fresnel(self, theory=None, **kw):
        for p, th in self.parts(theory):
            p.plot_fresnel(theory=th, **kw)

    plot_fresnel.__doc__ = Probe.plot_fresnel.__doc__

    def plot_logfresnel(self, theory=None, **kw):
        for p, th in self.parts(theory):
            p.plot_logfresnel(theory=th, **kw)

    plot_logfresnel.__doc__ = Probe.plot_logfresnel.__doc__

    def plot_Q4(self, theory=None, **kw):
        for p, th in self.parts(theory):
            p.plot_Q4(theory=th, **kw)

    plot_Q4.__doc__ = Probe.plot_Q4.__doc__

    def plot_residuals(self, theory=None, **kw):
        for p, th in self.parts(theory):
            p.plot_residuals(theory=th, **kw)

    plot_residuals.__doc__ = Probe.plot_residuals.__doc__

    def parts(self, theory):
        if theory is None:
            for p in self.probes:
                yield p, None
        else:
            offset = 0
            Q, R = theory
            for p in self.probes:
                n = len(p.Q)
                yield p, (Q[offset : offset + n], R[offset : offset + n])
                offset += n

    def shared_beam(self, intensity=1, background=0, back_absorption=1, theta_offset=0, sample_broadening=0):
        """
        Share beam parameters across all segments.

        New parameters are created for *intensity*, *background*,
        *theta_offset*, *sample_broadening* and *back_absorption*
        and assigned to the all segments.  These can be replaced
        with an explicit parameter in an individual segment if that
        parameter is independent.
        """
        intensity = Parameter.default(intensity, name="intensity")
        background = Parameter.default(background, name="background", limits=[0, None])
        back_absorption = Parameter.default(back_absorption, name="back_absorption", limits=[0, 1])
        theta_offset = Parameter.default(theta_offset, name="theta_offset")
        sample_broadening = Parameter.default(sample_broadening, name="sample_broadening", limits=[None, None])
        for p in self.probes:
            p.intensity = intensity
            p.background = background
            p.back_absorption = back_absorption
            p.theta_offset = theta_offset
            p.sample_broadening = sample_broadening

    def stitch(self, same_Q=0.001, same_dQ=0.001):
        r"""
        Stitch together multiple datasets into a single dataset.

        Points within *tol* of each other and with the same resolution
        are combined by interpolating them to a common $Q$ value then averaged
        using Gaussian error propagation.

        :Returns: probe | Probe
            Combined data set.

        :Algorithm:

        To interpolate a set of points to a common value, first find the
        common $Q$ value:

        .. math::

            \hat Q = \sum{Q_k} / n

        Then for each dataset $k$, find the interval $[i, i+1]$ containing the
        value $Q$, and use it to compute interpolated value for $R$:

        .. math::

            w &= (\hat Q - Q_i)/(Q_{i+1} - Q_i) \\
            \hat R &= w R_{i+1} + (1-w) R_{i+1} \\
            \hat \sigma_{R} &=
                \sqrt{ w^2 \sigma^2_{R_i} + (1-w)^2 \sigma^2_{R_{i+1}} } / n

        Average the resulting $R$ using Gaussian error propagation:

        .. math::

            \hat R &= \sum{\hat R_k}/n \\
            \hat \sigma_R &= \sqrt{\sum \hat \sigma_{R_k}^2}/n

        """
        Q, dQ, R, dR = stitch(self.probes)
        Po = self.probes[0]
        return QProbe(
            Q,
            dQ,
            data=(R, dR),
            intensity=Po.intensity,
            background=Po.background,
            back_absorption=Po.back_absorption,
            back_reflectivity=Po.back_reflectivity,
            resolution=Po.resolution,
        )


def parse_orso(filename, *args, **kwargs):
    """load an ORSO text (.ort) or binary (.orb) file"""
    if filename.endswith(".ort"):
        from orsopy.fileio.orso import load_orso

        entries = load_orso(filename)
    elif filename.endswith(".orb"):
        from orsopy.fileio.orso import load_nexus

        entries = load_nexus(filename)

    POL_CONVERSION = {
        "po": "++",
        "mo": "--",
        "mm": "--",
        "mp": "-+",
        "pm": "+-",
        "pp": "++",
    }

    entries_out = []
    for entry in entries:
        header = entry.info
        data = entry.data
        settings = header.data_source.measurement.instrument_settings
        columns = header.columns
        polarization = POL_CONVERSION.get(settings.polarization, "unpolarized")
        header_out = {"polarization": polarization}

        def get_key(orso_name, refl1d_name, refl1d_resolution_name):
            column_index = next(
                (i for i, c in enumerate(columns) if getattr(c, "physical_quantity", None) == orso_name), None
            )
            if column_index is not None:
                # NOTE: this is based on column being second index (under debate in ORSO)
                header_out[refl1d_name] = data[:, column_index]
                cname = columns[column_index].name
                resolution_index = next(
                    (i for i, c in enumerate(columns) if getattr(c, "error_of", None) == cname), None
                )
                if resolution_index is not None:
                    header_out[refl1d_resolution_name] = data[:, resolution_index]
            else:
                v = getattr(settings, orso_name, None)
                if hasattr(v, "magnitude"):
                    header_out[refl1d_name] = v.magnitude
                if hasattr(v, "error"):
                    header_out[refl1d_resolution_name] = v.error.error_value

        get_key("incident_angle", "angle", "angular_resolution")
        get_key("wavelength", "wavelength", "wavelength_resolution")

        entries_out.append((header_out, np.array(data).T))
    return entries_out


def load4(
    filename,
    keysep=":",
    sep=None,
    comment="#",
    name=None,
    intensity=1,
    background=0,
    back_absorption=1,
    back_reflectivity=False,
    Aguide=BASE_GUIDE_ANGLE,
    H=0,
    theta_offset=None,
    sample_broadening=None,
    L=None,
    dL=None,
    T=None,
    dT=None,
    dR=None,
    FWHM=False,
    radiation=None,
    columns=None,
    data_range=(None, None),
    resolution="normal",
    oversampling=None,
):
    r"""
    Load in four column data Q, R, dR, dQ.

    The file is loaded with *bumps.data.parse_multi*.  *keysep*
    defaults to ':' so that header data looks like JSON key: value
    pairs.  *sep* is None so that the data uses white-space separated
    columns.  *comment* is the standard '#' comment character, used
    for "# key: value" lines, for commenting out data lines using
    "#number number number number", and for adding comments after
    individual data lines.  The parser isn't very sophisticated, so
    be nice.

    *intensity* is the overall beam intensity, *background* is the
    overall background level, and *back_absorption* is the relative
    intensity of data measured at negative Q compared to positive Q
    data.  These can be values or a bumps *Parameter* objects.

    *back_reflectivity* is True if reflectivity was measured through
    the substrate.  This allows you to arrange the model from substrate
    to surface regardless of whether you are measuring through the
    substrate or reflecting off the surface.

    *theta_offset* indicates sample alignment.  In order to use theta
    offset you need to be able to convert from Q to wavelength and angle
    by providing values for the wavelength or the angle, and the associated
    resolution.

    *L*, *dL* in Angstroms can be used to recover angle and angular resolution
    for monochromatic sources where wavelength is fixed and angle is varying.
    These values can also be stored in the file header as::

        # wavelength: 4.75  # Ang
        # wavelength_resolution: 0.02  # Ang (1-sigma)

    *T*, *dT* in degrees can be used to recover wavelength and wavelength
    dispersion for time of flight sources where angle is fixed and wavelength
    is varying, or you can store them in the header of the file::

        # angle: 2  # degrees
        # angular_resolution: 0.2  # degrees (1-sigma)

    If both angle and wavelength are varying in the data, you can specify
    a separate value for each point, such the following::

        # wavelength: [1, 1.2, 1.5, 2.0, ...]
        # wavelength_resolution: [0.02, 0.02, 0.02, ...]

    *dR* can be used to replace the uncertainty estimate for R in the
    file with $\Delta R = R * \text{dR}$.  This allows files with only
    two columns, *Q* and *R* to be loaded.  Note that points with *dR=0*
    are automatically set to the minimum *dR>0* in the dataset.

    Instead of constants, you can provide function, *dT = lambda T: f(T)*,
    *dL = lambda L: f(L)* or *dR = lambda Q, R, dR: f(Q, R, dR)* for more
    complex relationships (with *dR()* returning 1-$\sigma$ $\Delta R$).

    *sample_broadening* in degrees FWHM adds to the angular_resolution.
    Scale 1-$\sigma$ rms by $2 \surd(2 \ln 2) \approx 2.34$ to convert to FWHM.

    *Aguide* and *H* are parameters for polarized beam measurements
    indicating the magnitude and direction of the applied field.

    Polarized data is represented using a multi-section data file,
    with blank lines separating each section.  Each section must
    have a polarization keyword, with value "++", "+-", "-+" or "--".

    *FWHM* is True if dQ, dT, dL are given as FWHM rather than 1-$\sigma$.
    *dR* is always 1-$\sigma$.  *sample_broadening* is always FWHM.

    *radiation* is 'xray' or 'neutron', depending on whether X-ray or
    neutron scattering length density calculator should be used for
    determining the scattering length density of a material.
    Default is 'neutron'

    *columns* is a string giving the column order in the file.  Default
    order is "Q R dR dQ".  Note: include dR and dQ even if the file only
    has two or three columns, but put the missing columns at the end.

    *data_range* indicates which data rows to use.  Arguments are the
    same as the list slice arguments, *(start, stop, step)*.  This follows
    the usual semantics of list slicing, *L[start:stop:step]*, with
    0-origin indices, stop is last plus one and step optional.  Use negative
    numbers to count from the end.  Default is *(None, None)* for the entire
    data set.

    *resolution* is 'normal' (default) or 'uniform'. Use uniform if you
    are merging Q points from a finely stepped energy sensitive measurement.

    *oversampling* is None or a positive integer indicating how many points to add
    between data point to support sparse data with denser theory (for PolarizedNeutronProbe)
    """
    json_header_encoding = False

    if filename.endswith(".ort") or filename.endswith(".orb"):
        entries = parse_orso(filename)
    else:
        json_header_encoding = True  # for .refl files, header values are json-encoded
        entries = parse_multi(filename, keysep=keysep, sep=sep, comment=comment)

    if columns:
        actual = columns.split()
        natural = "Q R dR dQ".split()
        column_order = [actual.index(k) for k in natural]
    else:
        column_order = [0, 1, 2, 3]
    index = slice(*data_range)
    probe_args = dict(
        name=name,
        filename=filename,
        intensity=intensity,
        background=background,
        back_absorption=back_absorption,
        back_reflectivity=back_reflectivity,
        theta_offset=theta_offset,
        sample_broadening=sample_broadening,
        resolution=resolution,
    )
    data_args = dict(
        radiation=radiation,
        FWHM=FWHM,
        T=T,
        L=L,
        dT=dT,
        dL=dL,
        dR=dR,
        column_order=column_order,
        index=index,
    )
    if len(entries) == 1:
        probe = _data_as_probe(entries[0], json_header_encoding, probe_args, **data_args)
    else:
        data_by_xs = {
            strip_quotes(entry[0]["polarization"]): _data_as_probe(entry, json_header_encoding, probe_args, **data_args)
            for entry in entries
        }
        if not set(data_by_xs.keys()) <= set("-- -+ +- ++".split()):
            raise ValueError("Unknown cross sections in: " + ", ".join(sorted(data_by_xs.keys())))
        xs = [data_by_xs.get(xs, None) for xs in ("--", "-+", "+-", "++")]

        if any(isinstance(d, QProbe) for d in xs if d is not None):
            probe = PolarizedQProbe(xs, Aguide=Aguide, H=H)
        else:
            probe = PolarizedNeutronProbe(xs, Aguide=Aguide, H=H, oversampling=oversampling)
    return probe


def _data_as_probe(entry, json_header_encoding, probe_args, T, L, dT, dL, dR, FWHM, radiation, column_order, index):
    decoder = json.loads if json_header_encoding else lambda x: x
    name = probe_args["filename"]
    header, data = entry
    if len(data) == 2:
        data_Q, data_R = (data[k][index] for k in column_order[:2])
        data_dR, data_dQ = None, None
        if dR is None:
            raise ValueError("Need dR for two column data in %r" % name)
    elif len(data) == 3:
        data_Q, data_R, data_dR = (data[k][index] for k in column_order[:3])
        data_dQ = None
    else:
        data_Q, data_R, data_dR, data_dQ = (data[k][index] for k in column_order)

    if FWHM and data_dQ is not None:  # dQ is already 1-sigma when not FWHM
        data_dQ = FWHM2sigma(data_dQ)

    # Override dR in the file if desired.
    # Make sure the computed dR is positive (otherwise chisq is infinite) by
    # choosing the smallest positive uncertainty to replace the invalid values.
    if dR is not None:
        data_dR = dR(data_Q, data_R, data_dR) if callable(dR) else data_R * dR
        data_dR[data_dR <= 0] = np.min(data_dR[data_dR > 0])

    # support calculation of sld from material based on radiation type
    if radiation is not None:
        data_radiation = radiation
    elif "radiation" in header:
        data_radiation = decoder(header["radiation"])
    else:
        # Default to neutron data if radiation not given in head.
        data_radiation = "neutron"
        # data_radiation = None

    if data_radiation == "xray":
        make_probe = XrayProbe
    elif data_radiation == "neutron":
        make_probe = NeutronProbe
    else:
        make_probe = Probe

    # Get T,dT,L,dL from header if it is not provided as an argument
    def fetch_key(key, override):
        # Note: pulls header and index pulled from context
        if override is not None:
            return override
        elif key in header:
            v = decoder(header[key])
            return np.array(v)[index] if isinstance(v, list) else v
        else:
            return None

    # Get T and L, either from user input or from datafile.
    data_T = fetch_key("angle", T)
    data_L = fetch_key("wavelength", L)

    # If one of T and L is missing, reconstruct it from Q
    if data_T is None and data_L is not None:
        data_T = QL2T(data_Q, data_L)
    if data_L is None and data_T is not None:
        data_L = QT2L(data_Q, data_T)

    # Get dT and dL, either from user input or from datafile.
    data_dL = fetch_key("wavelength_resolution", dL)
    data_dT = fetch_key("angular_resolution", dT)
    # print(header['angular_resolution'], data_dT)

    # Support dT = f(T), dL = f(L)
    if callable(data_dT):
        if data_T is None:
            raise ValueError("Need T to determine dT for %r" % name)
        data_dT = data_dT(data_T)
    if callable(data_dL):
        if data_L is None:
            raise ValueError("Need L to determine dL for %r" % name)
        data_dL = data_dL(data_L)

    # Convert input dT,dL to FWHM if necessary.
    if data_dL is not None and not FWHM:
        data_dL = sigma2FWHM(data_dL)
    if data_dT is not None and not FWHM:
        data_dT = sigma2FWHM(data_dT)

    # If dT or dL is missing, reconstruct it from Q.
    if data_dT is None and not any(v is None for v in (data_L, data_dL, data_dQ)):
        data_dT = dQdL2dT(data_Q, data_dQ, data_L, data_dL)
    if data_dL is None and not any(v is None for v in (data_T, data_dT, data_dQ)):
        data_dLoL = dQdT2dLoL(data_Q, data_dQ, data_T, data_dT)
        data_dL = data_dLoL * data_L

    # Check reconstruction if user provided any of T, L, dT, or dL.
    # Also, sample_offset or sample_broadening.
    offset = probe_args["theta_offset"]
    broadening = probe_args["sample_broadening"]
    if any(v is not None for v in (T, dT, L, dL, offset, broadening)):
        if data_T is None:
            raise ValueError("Need L to determine T from Q for %r" % name)
        if data_L is None:
            raise ValueError("Need T to determine L from Q for %r" % name)
        if data_dT is None:
            raise ValueError("Need dL to determine dT from dQ for %r" % name)
        if data_dL is None:
            raise ValueError("Need dT to determine dL from dQ for %r" % name)

    # Build the probe, or the Q probe if we don't have angle and wavelength.
    if all(v is not None for v in (data_T, data_L, data_dT, data_dL)):
        probe = make_probe(
            T=data_T, dT=data_dT, L=data_L, dL=data_dL, data=(data_R, data_dR), dQo=data_dQ, **probe_args
        )
    else:
        # QProbe doesn't accept theta_offset or sample_broadening
        qprobe_args = probe_args.copy()
        qprobe_args.pop("theta_offset")
        qprobe_args.pop("sample_broadening")
        probe = QProbe(data_Q, data_dQ, data=(data_R, data_dR), **qprobe_args)

    return probe


@dataclass(init=False)
class QProbe(BaseProbe):
    """
    A pure Q, R probe

    This probe with no possibility of tricks such as looking up the
    scattering length density based on wavelength, or adjusting for
    alignment errors.
    """

    Q: "NDArray"
    dQ: "NDArray"
    name: Optional[str]
    filename: Optional[str]
    intensity: Parameter
    back_absorption: Parameter
    background: Parameter
    back_reflectivity: bool
    R: "NDArray"
    dR: "NDArray"
    resolution: Literal["normal", "uniform"]

    polarized = False

    @classmethod
    def from_dict(cls, **kw):
        R = kw.pop("R", None)
        dR = kw.pop("dR", None)
        if R is not None and dR is not None:
            kw["data"] = (R, dR)
        return cls(**kw)

    def __init__(
        self,
        Q,
        dQ,
        data=None,
        name=None,
        filename=None,
        intensity=1,
        background=0,
        back_absorption=1,
        back_reflectivity=False,
        resolution: Literal["normal", "uniform"] = "normal",
    ):
        if not name and filename:
            name = os.path.splitext(os.path.basename(filename))[0]
        qualifier = " " + name if name is not None else ""
        self.intensity = Parameter.default(intensity, name="intensity" + qualifier)
        self.background = Parameter.default(background, name="background" + qualifier, limits=[0, None])
        self.back_absorption = Parameter.default(back_absorption, name="back_absorption" + qualifier, limits=[0, 1])

        self.back_reflectivity = back_reflectivity

        if data is not None:
            R, dR = data
        else:
            R, dR = None, None

        self.Q, self.dQ = Q, dQ
        self.R = R
        self.dR = dR
        self.unique_L = None
        self.calc_Qo = self.Q
        self.name = name
        self.filename = filename
        self.resolution = resolution

    @property
    def calc_Q(self):
        return self.calc_Qo if not self.back_reflectivity else -self.calc_Qo

    def parameters(self):
        return {
            "intensity": self.intensity,
            "background": self.background,
            "back_absorption": self.back_absorption,
        }

    def scattering_factors(self, material, density):
        raise NotImplementedError(
            "need radiation type and wavelength in <%s> to compute sld for %s" % (self.filename, material)
        )

    scattering_factors.__doc__ = Probe.scattering_factors.__doc__

    def oversample(self, n=20, seed=1):
        rng = numpy.random.RandomState(seed=seed)
        extra = rng.normal(self.Q, self.dQ, size=(n - 1, len(self.Q)))
        calc_Q = np.hstack((self.Q, extra.flatten()))
        self.calc_Qo = np.sort(calc_Q)

    oversample.__doc__ = Probe.oversample.__doc__

    def critical_edge(self, substrate=None, surface=None, n=51, delta=0.25):
        Q_c = self.Q_c(substrate, surface)
        extra = np.linspace(Q_c * (1 - delta), Q_c * (1 + delta), n)
        calc_Q = np.hstack((self.Q, extra, 0))
        self.calc_Qo = np.sort(calc_Q)

    critical_edge.__doc__ = Probe.critical_edge.__doc__


def measurement_union(xs):
    """
    Determine the unique (T, dT, L, dL) across all datasets.
    """

    # First gather a set of unique tuples in angle and wavelength
    TL = set()
    for x in xs:
        if x is not None:
            TL |= set(zip(x.calc_T, x.calc_T + x.theta_offset.value, x.calc_L))
    To, T, L = zip(*[item for item in TL])
    To, T, L = [np.asarray(v) for v in (To, T, L)]

    # Convert to Q, dQ
    Q = TL2Q(T, L)

    # Sort by Q
    idx = np.argsort(Q)
    To, T, L, Q = To[idx], T[idx], L[idx], Q[idx]
    if abs(Q[1:] - Q[:-1]).any() < 1e-14:
        raise ValueError("Q is not unique")
    return To, T, L, Q


def Qmeasurement_union(xs):
    """
    Determine the unique Q, dQ across all datasets.
    """
    Qset = set()
    for x in xs:
        if x is not None:
            Qset |= set(zip(x.Q, x.dQ))
    Q, dQ = [np.array(v) for v in zip(*sorted(Qset))]
    if abs(Q[1:] - Q[:-1]).any() < 1e-14:
        raise ValueError("Q values differ by less than 1e-14")
    return Q, dQ


optional_xs = Union[NeutronProbe, Literal[None]]


@dataclass(init=False)
class PolarizedNeutronProbe:
    """
    Polarized neutron probe

    *xs* (4 x NeutronProbe) is a sequence pp, pm, mp and mm.

    *Aguide* (degrees) is the angle of the applied field relative
    to the plane of the sample, with angle 270º in the plane of the sample.

    *H* (tesla) is the magnitude of the applied field
    """

    name: str
    mm: optional_xs = None
    mp: optional_xs = None
    pm: optional_xs = None
    pp: optional_xs = None
    H: Parameter
    Aguide: Parameter
    oversampling: Optional[int] = None
    oversampling_seed: int = 1

    view = None  # Default to Probe.view when None
    show_resolution = None  # Default to Probe.show_resolution when None
    substrate = surface = None
    polarized = True
    _xs_names = ["mm", "mp", "pm", "pp"]

    def __init__(
        self,
        xs: Optional[Tuple] = None,
        mm: optional_xs = None,
        mp: optional_xs = None,
        pm: optional_xs = None,
        pp: optional_xs = None,
        name=None,
        Aguide=BASE_GUIDE_ANGLE,
        H=0,
        oversampling=None,
        oversampling_seed=1,
    ):
        if any([mm, mp, pm, pp]):
            if xs is not None:
                warnings.warn("a cross-section is directly specified - xs argument will be ignored")
            self.mm = mm
            self.mp = mp
            self.pm = pm
            self.pp = pp
        else:
            for index, xs_name in enumerate(self._xs_names):
                setattr(self, xs_name, xs[index])

        self._theta_offsets = None
        if name is None and self.mm is not None:
            name = self.mm.name
        self.name = name
        self._calculate_union()
        # self.T, self.dT, self.L, self.dL, self.Q, self.dQ \
        #     = measurement_union(xs)
        # self._set_calc(self.T, self.L)
        self._check()
        spec = " " + name if name else ""
        self.H = Parameter.default(H, name="H" + spec)
        self.Aguide = Parameter.default(Aguide, name="Aguide" + spec, limits=[-360, 360])
        if oversampling is not None:
            self.oversample(oversampling, oversampling_seed)

    @property
    def xs(self) -> List[Union[NeutronProbe, None]]:
        return [getattr(self, xs_name) for xs_name in self._xs_names]

    def parameters(self):
        xs_pars = [(xsi.parameters() if xsi else None) for xsi in self.xs]
        output = dict(zip(self._xs_names, xs_pars))
        output["Aguide"] = self.Aguide
        output["H"] = self.H
        return output

    def resynth_data(self):
        for p in self.xs:
            if p is not None:
                p.resynth_data()

    resynth_data.__doc__ = Probe.resynth_data.__doc__

    def restore_data(self):
        for p in self.xs:
            if p is not None:
                p.restore_data()

    restore_data.__doc__ = Probe.restore_data.__doc__

    def simulate_data(self, theory, noise=2.0):
        if noise is None or np.isscalar(noise):
            noise = [noise] * 4
        for data_k, theory_k, noise_k in zip(self.xs, theory, noise):
            if data_k is not None:
                data_k.simulate_data(theory=theory_k, noise=noise_k)

    simulate_data.__doc__ = Probe.simulate_data.__doc__

    def _check(self):
        back_refls = [f.back_reflectivity for f in self.xs if f is not None]
        if all(back_refls) or not any(back_refls):
            self.back_reflectivity = back_refls[0]
        else:
            raise ValueError("Cannot mix front and back reflectivity measurements")

    def shared_beam(self, intensity=1, background=0, back_absorption=1, theta_offset=0, sample_broadening=0):
        """
        Share beam parameters across all four cross sections.

        New parameters are created for *intensity*, *background*,
        *theta_offset*, *sample_broadening* and *back_absorption*
        and assigned to the all cross sections.  These can be replaced
        with an explicit parameter in an individual cross section if that
        parameter is independent.
        """
        intensity = Parameter.default(intensity, name="intensity")
        background = Parameter.default(background, name="background", limits=[0, None])
        back_absorption = Parameter.default(back_absorption, name="back_absorption", limits=[0, 1])
        theta_offset = Parameter.default(theta_offset, name="theta_offset")
        sample_broadening = Parameter.default(sample_broadening, name="sample_broadening", limits=[None, None])
        for x in self.xs:
            if x is not None:
                x.intensity = intensity
                x.background = background
                x.back_absorption = back_absorption
                x.theta_offset = theta_offset
                x.sample_broadening = sample_broadening

    def oversample(self, n, seed=1):
        for xs in self.xs:
            if xs is not None:
                xs.oversample(n, seed=1)
        self._theta_offsets = None

    oversample.__doc__ = Probe.oversample.__doc__

    def _calculate_union(self):
        theta_offsets = [x.theta_offset.value for x in self.xs if x is not None]
        if self._theta_offsets is not None and theta_offsets == self._theta_offsets:
            # no change in offsets: use cached values of measurement union
            return

        unique_offsets = set(theta_offsets)
        shared_offset = theta_offsets[0] if len(unique_offsets) == 1 else None
        if shared_offset is not None and self._theta_offsets is not None:
            # offset is shared, and measurement_union has been calculated before
            # (so self.To is already populated)
            self.T = self._To + shared_offset
            Q = TL2Q(T=self.T, L=self.L)
            self.calc_Qo = Q
        else:
            # unshared offsets changed, or union has not been calculated before
            # returned T has offset applied, To does not
            self._To, self.T, self.L, self.Q = measurement_union(self.xs)
            self._set_calc(self.T, self.L)

        self._theta_offsets = theta_offsets

    @property
    def calc_Q(self):
        # print('calculating calc_Q...')
        self._calculate_union()
        return self.calc_Qo if not self.back_reflectivity else -self.calc_Qo

    def _set_calc(self, T, L):
        # TODO: shouldn't clone code from probe
        Q = TL2Q(T=T, L=L)

        idx = np.argsort(Q)
        self.calc_T = T[idx]
        self.calc_L = L[idx]
        self.calc_Qo = Q[idx]

        # Only keep the scattering factors that you need
        self.unique_L = np.unique(self.calc_L)
        self._L_idx = np.searchsorted(self.unique_L, L)

    def apply_beam(self, Q, R, resolution=True, interpolation=0):
        """
        Apply factors such as beam intensity, background, backabsorption,
        and footprint to the data.
        """
        return [(xs.apply_beam(Q, Ri, resolution, interpolation) if xs else None) for xs, Ri in zip(self.xs, R)]

    def fresnel(self, *args, **kw):
        return self.pp.fresnel(*args, **kw)

    fresnel.__doc__ = Probe.fresnel.__doc__

    def scattering_factors(self, material, density):
        # doc string is inherited from parent (see below)
        rho, irho, rho_incoh = nsf.neutron_sld(material, wavelength=self.unique_L[0], density=density)
        # TODO: support wavelength dependent systems
        # print("sf", str(material), type(rho), type(irho[0]))
        return rho, irho, rho_incoh
        # return rho, irho[self._L_idx], rho_incoh

    scattering_factors.__doc__ = Probe.scattering_factors.__doc__

    def select_corresponding(self, theory):
        """
        Select theory points corresponding to the measured data.

        Since we have evaluated theory at every Q, it is safe to interpolate
        measured Q into theory, since it will land on a node,
        not in an interval.
        """

        Qth, Rth = theory
        return [
            None if x_data is None else (x_data.Q, np.interp(x_data.Q, Qth, x_th)) for x_data, x_th in zip(self.xs, Rth)
        ]

    def save(self, filename, theory, substrate=None, surface=None):
        for xsi, xsi_th, suffix in zip(self.xs, theory, ("A", "B", "C", "D")):
            if xsi is not None:
                xsi.save(filename + suffix, xsi_th, substrate=substrate, surface=surface)

    save.__doc__ = Probe.save.__doc__

    def plot(self, view=None, **kwargs):
        """
        Plot theory against data.

        Need substrate/surface for Fresnel-normalized reflectivity
        """
        view = view if view is not None else self.view

        if view is None:
            view = Probe.view  # Default to Probe.view

        if view == "linear":
            self.plot_linear(**kwargs)
        elif view == "log":
            self.plot_log(**kwargs)
        elif view == "fresnel":
            self.plot_fresnel(**kwargs)
        elif view == "logfresnel":
            self.plot_logfresnel(**kwargs)
        elif view == "q4":
            self.plot_Q4(**kwargs)
        elif view.startswith("resid"):
            self.plot_residuals(**kwargs)
        elif view == "SA":
            self.plot_SA(**kwargs)
        elif view == "resolution":
            self.plot_resolution(**kwargs)
        else:
            raise TypeError("incorrect reflectivity view '%s'" % view)

    def plot_resolution(self, **kwargs):
        self._xs_plot("plot_resolution", **kwargs)

    def plot_linear(self, **kwargs):
        self._xs_plot("plot_linear", **kwargs)

    def plot_log(self, **kwargs):
        self._xs_plot("plot_log", **kwargs)

    def plot_fresnel(self, **kwargs):
        self._xs_plot("plot_fresnel", **kwargs)

    def plot_logfresnel(self, **kwargs):
        self._xs_plot("plot_logfresnel", **kwargs)

    def plot_Q4(self, **kwargs):
        self._xs_plot("plot_Q4", **kwargs)

    def plot_residuals(self, **kwargs):
        self._xs_plot("plot_residuals", **kwargs)

    def plot_SA(self, theory=None, label=None, plot_shift=None, **kwargs):
        import matplotlib.pyplot as plt

        if self.pp is None or self.mm is None:
            raise TypeError("cannot form spin asymmetry plot without ++ and --")

        plot_shift = plot_shift if plot_shift is not None else Probe.plot_shift
        trans = auto_shift(plot_shift)
        pp, mm = self.pp, self.mm
        c = coordinated_colors()
        if hasattr(pp, "R") and hasattr(mm, "R") and pp.R is not None and mm.R is not None:
            Q, SA, dSA = spin_asymmetry(pp.Q, pp.R, pp.dR, mm.Q, mm.R, mm.dR)
            if dSA is not None:
                res = self.show_resolution if self.show_resolution is not None else Probe.show_resolution
                dQ = pp.dQ if res else None
                plt.errorbar(
                    Q,
                    SA,
                    yerr=dSA,
                    xerr=dQ,
                    fmt=".",
                    capsize=0,
                    label=pp.label(prefix=label, gloss="data"),
                    transform=trans,
                    color=c["light"],
                )
            else:
                plt.plot(Q, SA, ".", label=pp.label(prefix=label, gloss="data"), transform=trans, color=c["light"])
            # Set limits based on max theoretical SA, which is in (-1.0, 1.0)
            # If the error bars are bigger than that, you usually don't care.
            ylim_low, ylim_high = plt.ylim()
            plt.ylim(max(ylim_low, -2.5), min(ylim_high, 2.5))
        if theory is not None:
            mm, mp, pm, pp = theory
            Q, SA, _ = spin_asymmetry(pp[0], pp[1], None, mm[0], mm[1], None)
            plt.plot(Q, SA, label=self.pp.label(prefix=label, gloss="theory"), transform=trans, color=c["dark"])
        plt.xlabel(r"Q (\AA^{-1})")
        plt.ylabel(r"spin asymmetry $(R^{++} -\, R^{--}) / (R^{++} +\, R^{--})$")
        plt.legend(numpoints=1)

    def _xs_plot(self, plotter, theory=None, **kwargs):
        import matplotlib.pyplot as plt

        # Plot available cross sections
        if theory is None:
            theory = (None, None, None, None)
        for x_data, x_th, suffix in zip(self.xs, theory, ("$^{--}$", "$^{-+}$", "$^{+-}$", "$^{++}$")):
            if x_data is not None:
                fn = getattr(x_data, plotter)
                fn(theory=x_th, suffix=suffix, **kwargs)


def spin_asymmetry(Qp, Rp, dRp, Qm, Rm, dRm):
    r"""
    Compute spin asymmetry for R++, R--.

    **Parameters:**

    *Qp*, *Rp*, *dRp* : vector
        Measured ++ cross section and uncertainty.
    *Qm*, *Rm*, *dRm* : vector
        Measured -- cross section and uncertainty.

    If *dRp*, *dRm* are None then the returned uncertainty will also be None.

    **Returns:**

    *Q*, *SA*, *dSA* : vector
        Computed spin asymmetry and uncertainty.

    **Algorithm:**

    Spin asymmetry, $S_A$, is:

    .. math::

        S_A = \frac{R_{++} - R_{--}}{R_{++} + R_{--}}

    Uncertainty $\Delta S_A$ follows from propagation of error:

    .. math::

        \Delta S_A^2 = \frac{4(R_{++}^2\Delta R_{--}^2+R_{--}^2\Delta R_{++})}
                            {(R_{++} + R_{--})^4}

    """
    Rm = np.interp(Qp, Qm, Rm)
    v = (Rp - Rm) / (Rp + Rm)
    if dRp is not None:
        dRm = np.interp(Qp, Qm, dRm)
        dvsq = 4 * ((Rp * dRm) ** 2 + (Rm * dRp) ** 2) / (Rp + Rm) ** 4
        dvsq[dvsq < 0] = 0
        return Qp, v, sqrt(dvsq)
    else:
        return Qp, v, None


def _interpolate_Q(Q, dQ, n):
    """
    Helper function to interpolate between data points.

    *n* is the number of points to show between existing points.
    """
    if n > 0:
        # Extend the Q-range by 1/2 interval on either side
        Q = np.hstack((0.5 * (3.0 * Q[0] - Q[1]), Q, 0.5 * (3.0 * Q[-1] - Q[-2])))
        dQ = np.hstack((0.5 * (3.0 * dQ[0] - dQ[1]), dQ, 0.5 * (3.0 * dQ[-1] - dQ[-2])))
        index = np.arange(0, len(Q), dtype="d")
        subindex = np.arange(0, (n + 1) * (len(Q) - 1) + 1, dtype="d") / (n + 1.0)
        Q = np.interp(subindex, index, Q)
        dQ = np.interp(subindex, index, dQ)
    return Q, dQ


@dataclass(init=False)
class PolarizedQProbe(PolarizedNeutronProbe):
    polarized = True

    def __init__(
        self,
        xs: Optional[Tuple] = None,
        mm: optional_xs = None,
        mp: optional_xs = None,
        pm: optional_xs = None,
        pp: optional_xs = None,
        name=None,
        Aguide=BASE_GUIDE_ANGLE,
        H=0,
        oversampling: Optional[int] = None,
        oversampling_seed: int = 1,
    ):
        if any([mm, mp, pm, pp]):
            if xs is not None:
                warnings.warn("a cross-section is directly specified - xs argument will be ignored")
            self.mm = mm
            self.mp = mp
            self.pm = pm
            self.pp = pp
        else:
            for index, xs_name in enumerate(self._xs_names):
                setattr(self, xs_name, xs[index])
        self._check()
        self.name = name if name is not None else self.pp.name
        self.unique_L = None
        self.Aguide = Parameter.default(Aguide, name="Aguide " + self.name, limits=[-360, 360])
        self.H = Parameter.default(H, name="H " + self.name)
        if oversampling is not None:
            self.oversample(oversampling, oversampling_seed)
        self.Q, self.dQ = Qmeasurement_union(self.xs)
        self.calc_Qo = self.Q

    @property
    def calc_Q(self):
        return self.calc_Qo if not self.back_reflectivity else -self.calc_Qo


# Deprecated old long name
PolarizedNeutronQProbe = PolarizedQProbe



"""
1-D and 2-D rebinning code.
"""

__all__ = ["bin_edges", "logbin_edges", "rebin", "rebin2d"]

import numpy as np


def bin_edges(C):
    r"""
    Construct bin edges *E* from equally spaced bin centers *C*.

    Assumes the edges lie half way between the centers.  This will only
    be true if the centers are evenly spaced, but may be good enough for
    visualization purposes even when they are not.  Ideally analysis would
    be performed on the raw data without rebinning.
    """
    E = 0.5 * (C[:-1] + C[1:])
    return np.hstack((C[0] - (E[0] - C[0]), E, C[-1] + (C[-1] - E[-1])))


def logbin_edges(L):
    r"""
    Construct bin edges *E* from logarithmically spaced bin centers *L*.

    Assuming fixed $\omega = \Delta\lambda/\lambda$ in the bins, the
    edges will be spaced logarithmically at:

    .. math::

        E_0     &= \min \lambda \\
        E_{i+1} &= E_i + \omega E_i = E_i (1+\omega)

    with centers $L$ half way between the edges:

    .. math::

        L_i = (E_i+E_{i+1})/2
            = (E_i + E_i (1+\omega))/2
            = E_i (2 + \omega)/2

    Solving for $E_i$, we can recover the edges from the centers:

    .. math::

        E_i = L_i \frac{2}{2+\omega}

    The final edge, $E_{n+1}$, does not have a corresponding center
    $L_{n+1}$ so we must determine it from the previous edge $E_n$:

    .. math::

        E_{n+1} = L_n \frac{2}{2+\omega}(1+\omega)

    The fixed $\omega$ can be retrieved from the ratio of any pair
    of bin centers using:

    .. math::

        \frac{L_{i+1}}{L_i} = \frac{ (E_{i+2}+E_{i+1})/2 }{ (E_{i+1}+E_i)/2 }
                          = \frac{ (E_{i+1}(1+\omega)+E_{i+1} }
                                  { (E_i(1+\omega)+E_i }
                          = \frac{E_{i+1}}{E_i}
                          = \frac{E_i(1+\omega)}{E_i} = 1 + \omega
    """
    if L[1] > L[0]:
        dLoL = L[1] / L[0] - 1
        last = 1 + dLoL
    else:
        dLoL = L[0] / L[1] - 1
        last = 1.0 / (1 + dLoL)
    E = L * 2 / (2 + dLoL)
    return np.hstack((E, E[-1] * last))


def rebin(x, I, xo, Io=None, dtype=np.float64):
    """
    Rebin a vector.

    x are the existing bin edges
    xo are the new bin edges
    I are the existing counts (one fewer than edges)

    Io will be used if present, but be sure that it is a contiguous
    array of the correct shape and size.

    dtype is the type to use for the intensity vectors.  This can be
    integer (uint8, uint16, uint32) or real (float32 or f, float64 or d).
    The edge vectors are all coerced to doubles.

    Note that total intensity is not preserved for integer rebinning.
    The algorithm uses truncation so total intensity will be down on
    average by half the total number of bins.
    """
    from .refllib import backend

    # Coerce axes to float arrays
    x, xo = _input(x, dtype="d"), _input(xo, dtype="d")
    shape_in = np.array([x.shape[0] - 1])
    shape_out = np.array([xo.shape[0] - 1])

    # Coerce counts to correct type and check shape
    if dtype is None:
        try:
            dtype = I.dtype
        except AttributeError:
            dtype = np.float64
    I = _input(I, dtype=dtype)
    if shape_in != I.shape:
        raise TypeError("input array incorrect shape %s" % I.shape)

    # Create output vector
    Io = _output(Io, shape_out, dtype=dtype)

    # Call rebin on type if it is available
    # try:
    #     rebincore = getattr(_cmodule, 'rebin_' + I.dtype.name)
    # except AttributeError:
    #     raise TypeError("rebin supports uint 8/16/32/64 and float 32/64, not "
    #                     + I.dtype.name)
    # rebincore(x, I, xo, Io)
    backend.rebin_counts(x, I, xo, Io)
    return Io


def rebin2d(x, y, I, xo, yo, Io=None, dtype=None):
    """
    Rebin a matrix.

    x, y are the existing bin edges
    xo, yo are the new bin edges
    I is the existing counts (one fewer than edges in each direction)

    For example, with x representing the column edges in each row and
    y representing the row edges in each column, the following
    represents a uniform field::

        >>> #from bumps.rebin import rebin2d
        >>> x, y = [0, 2, 4, 5], [0, 1, 3]
        >>> z = [[2, 2, 1], [4, 4, 2]]

    We can check this by rebinning with uniform size bins::

        >>> xo, yo = range(6), range(4)
        >>> zo = rebin2d(y, x, z, yo, xo)
        >>> bool((zo == 1.).all())
        True

    dtype is the type to use for the intensity vectors.  This can be
    integer (uint8, uint16, uint32) or real (float32 or f, float64 or d).
    The edge vectors are all coerced to doubles.

    Note that total intensity is not preserved for integer rebinning.
    The algorithm uses truncation so total intensity will be down on
    average by half the total number of bins.

    Io will be used if present, if it is contiguous and if it has the
    correct shape and type for the input.  Otherwise it will raise a
    TypeError.  This will allow you to rebin the slices of an appropriately
    ordered matrix without making copies.
    """
    from .refllib import backend

    # Coerce axes to float arrays
    x, y, xo, yo = [_input(v, dtype="d") for v in (x, y, xo, yo)]
    shape_in = np.array([x.shape[0] - 1, y.shape[0] - 1])
    shape_out = np.array([xo.shape[0] - 1, yo.shape[0] - 1])

    # Coerce counts to correct type and check shape
    if dtype is None:
        try:
            dtype = I.dtype
        except AttributeError:
            dtype = np.float64
    I = _input(I, dtype=dtype)
    if (shape_in != I.shape).any():
        raise TypeError("input array incorrect shape %s" % str(I.shape))

    # Create output vector
    Io = _output(Io, shape_out, dtype=dtype)

    # # Call rebin on type if it is available
    # try:
    #     rebincore = getattr(_cmodule, 'rebin2d_' + I.dtype.name)
    # except AttributeError:
    #     raise TypeError("rebin2d supports uint 8/16/32/64 and float 32/64, not "
    #                     + I.dtype.name)
    # print(x.shape, y.shape, I.shape, xo.shape, yo.shape, Io.shape)
    # print(x.dtype, y.dtype, I.dtype, xo.dtype, yo.dtype, Io.dtype)
    backend.rebin_counts_2D(x, y, I, xo, yo, Io)
    return Io


def _input(v, dtype="d"):
    """
    Force v to be a contiguous array of the correct type, avoiding copies
    if possible.
    """
    return np.ascontiguousarray(v, dtype=dtype)


def _output(v, shape, dtype=np.float64):
    """
    Create a contiguous array of the correct shape and type to hold a
    returned array, reusing an existing array if possible.
    """
    if v is None:
        return np.empty(shape, dtype=dtype)
    if not (
        isinstance(v, np.ndarray) and v.dtype == np.dtype(dtype) and (v.shape == shape).all() and v.flags.contiguous
    ):
        raise TypeError("output vector must be contiguous %s of size %s" % (dtype, shape))
    return v


# ================ Test code ==================
# TODO: move test code to its own file
def _check_one_1d(from_bins, val, to_bins, target):
    target = _input(target)
    for f, F in [(from_bins, val), (from_bins[::-1], val[::-1])]:
        for t, T in [(to_bins, target), (to_bins[::-1], target[::-1])]:
            result = rebin(f, F, t)
            assert np.linalg.norm(T - result) < 1e-14, "rebin failed for %s->%s %s" % (f, t, result)


def _check_all_1d():
    # Split a value
    _check_one_1d([1, 2, 3, 4], [10, 20, 30], [1, 2.5, 4], [20, 40])

    # bin is a superset of rebin
    _check_one_1d([0, 1, 2, 3, 4], [5, 10, 20, 30], [1, 2.5, 3], [20, 10])

    # bin is a subset of rebin
    _check_one_1d([1, 2, 3, 4, 5, 6], [10, 20, 30, 40, 50], [2.5, 3.5], [25])

    # one bin to many
    _check_one_1d([1, 2, 3, 4, 5, 6], [10, 20, 30, 40, 50], [2.1, 2.2, 2.3, 2.4], [2, 2, 2])

    # many bins to one
    _check_one_1d([1, 2, 3, 4, 5, 6], [10, 20, 30, 40, 50], [2.5, 4.5], [60])


def _check_one_2d(x, y, z, xo, yo, zo):
    # print("checking")
    # print(x, y, z)
    # print(xo, yo, zo)
    result = rebin2d(x, y, z, xo, yo)
    target = np.array(zo, dtype=result.dtype)
    assert np.linalg.norm(target - result) < 1e-14, "rebin2d failed for %s, %s->%s, %s\nexpected: %s\nbut got: %s" % (
        x,
        y,
        xo,
        yo,
        zo,
        z,
    )


def _check_uniform_2d(x, y):
    z = np.array([y], "d") * np.array([x], "d").T
    xedges = np.concatenate([(0,), np.cumsum(x)])
    yedges = np.concatenate([(0,), np.cumsum(y)])
    nx = int(np.round(xedges[-1]))
    ny = int(np.round(yedges[-1]))
    ox = np.arange(nx + 1)
    oy = np.arange(ny + 1)
    target = np.ones([nx, ny], "d")
    _check_one_2d(xedges, yedges, z, ox, oy, target)


def _check_all_2d():
    x, y, I = [0, 3, 5, 7], [0, 1, 3], [[3, 6], [2, 4], [2, 4]]
    xo, yo, Io = range(8), range(4), [[1] * 3] * 7
    x, y, I, xo, yo, Io = [np.array(A, "d") for A in (x, y, I, xo, yo, Io)]

    # Try various types and orders on a non-square matrix
    _check_one_2d(x, y, I, xo, yo, Io)
    _check_one_2d(x[::-1], y, I[::-1, :], xo, yo, Io)
    _check_one_2d(x, y[::-1], I[:, ::-1], xo, yo, Io)
    _check_one_2d(x, y, I, [7, 3, 0], yo, [[4] * 3, [3] * 3])
    _check_one_2d(x, y, I, xo, [3, 2, 0], [[1, 2]] * 7)
    _check_one_2d(y, x, I.T, yo, xo, Io.T)  # C vs. Fortran ordering

    # Test smallest possible result
    _check_one_2d([-1, 2, 4], [0, 1, 3], [[3, 6], [2, 4]], [1, 2], [1, 2], [1])
    # subset/superset
    _check_one_2d([0, 1, 2, 3], [0, 1, 2, 3], [[1] * 3] * 3, [0.5, 1.5, 2.5], [0.5, 1.5, 2.5], [[1] * 2] * 2)
    for dtype in ["uint8", "uint16", "uint32", "float32", "float64"]:
        _check_one_2d(
            [0, 1, 2, 3, 4],
            [0, 1, 2, 3, 4],
            np.array([[1] * 4] * 4, dtype=dtype),
            [-2, -1, 2, 5, 6],
            [-2, -1, 2, 5, 6],
            np.array([[0, 0, 0, 0], [0, 4, 4, 0], [0, 4, 4, 0], [0, 0, 0, 0]], dtype=dtype),
        )
    # non-square test
    _check_uniform_2d([1, 2.5, 4, 0.5], [3, 1, 2.5, 1, 3.5])
    _check_uniform_2d([3, 2], [1, 2])


def _check_bin_edges():
    log_edges = np.asarray([2 * 1.2**c for c in range(10)])
    log_centers = (log_edges[1:] + log_edges[:-1]) / 2
    assert np.linalg.norm(logbin_edges(log_centers) - log_edges) < 1e-10

    lin_edges = np.linspace(2, 10, 10)
    lin_centers = (lin_edges[1:] + lin_edges[:-1]) / 2
    assert np.linalg.norm(bin_edges(lin_centers) - lin_edges) < 1e-10


def test():
    _check_all_1d()
    _check_all_2d()
    _check_bin_edges()


if __name__ == "__main__":
    test()



"""
Biomolecule support.

:class:`Molecule` lets you define biomolecules with labile hydrogen atoms
specified using tritium (T) in the chemical formula.  The biomolecule object
creates forms with natural isotope ratio, all hydrogen and all deuterium.
Density can be provided as natural density or cell volume.  A %D2O contrast
match value is computed for matching the molecule SLD in the presence of
labile hydrogens.  :method:`Molecule.D2Osld` computes the neutron SLD for
the solvated molecule in a %D2O solvent.

:fun:`D2Omatch` computes the %D2O constrast match value given the fully
hydrogenated and fully deuterated forms.

:class:`Sequence` lets you read amino acid and DNA/RNA sequences from FASTA
files.

Tables for common molecules are provided[1]:

    *AMINO_ACID_CODES* : amino acids indexed by FASTA code

    *RNA_CODES*, DNA_CODES* : nucleic bases indexed by FASTA code

    *RNA_BASES*, DNA_BASES* : individual nucleic acid bases

    *NUCLEIC_ACID_COMPONENTS*, *LIPIDS*, *CARBOHYDRATE_RESIDUES*

Neutron SLD for water at 20C is also provided as *H2O_SLD* and *D2O_SLD*.

For unmodified protein need to add 2*T and O for terminations.

Assumes that proteins were created in an environment with the usual H/D isotope
ratio on the non-swappable hydrogens.

[1] Perkins, Modern Physical Methods in Biochemistry Part B, 143-265 (1988)

"""

from __future__ import division, print_function

import periodictable as pt


class Molecule(object):
    """
    Specify a biomolecule by name, chemical formula, cell volume and charge.

    Labile hydrogen positions should be coded using tritium (T) rather than H.  That
    way the tritium can be changed to H[1] for solutions with pure water, H for solutions
    with a natural abundance of water or D for solutions with pure deuterium.

    Attributes
    ==========

    *formula* is the original tritiated formula.  You can create the hydrogenated or
    deuterated forms using::

        from periodictable import H, D, T, fast
        hydrogenated = isotope_substitution
    :fun:`isotope_substitution(M.formula, ` with *formula*, T
    and periodictable.H or periodictable.D.

    *D2Omatch* is the % D2O in H2O required to contrast match the molecule, including
    the the proton swapping effect.

    *natural*/*H*/*D* are the formulae with tritium replaced by naturally occurring D/H ratio,
    by pure H and by pure D respectively.  *sld*/*H_sld*/*D_sld* are the corresponding slds.
    *mass* and *density* are available as *H.mass* and *H.density*, etc.

    *charge* is the charge on the molecule

    *cell_volume* is the estimated cell volume for the molecule
    """

    def __init__(self, name, formula, cell_volume=None, natural_density=None, charge=0):
        M = pt.formula(formula, natural_density=natural_density)

        # Fill in density or cell_volume
        if cell_volume is not None:
            M.density = 1e24 * M.molecular_mass / cell_volume if cell_volume > 0 else 0
        else:
            cell_volume = 1e24 * M.molecular_mass / M.density

        self.cell_volume = cell_volume
        self.name = name
        self.formula = M
        self.natural = isotope_substitution(M, pt.T, pt.H)
        self.H = isotope_substitution(M, pt.T, pt.H[1])
        self.D = isotope_substitution(M, pt.T, pt.D)

        self.sld = pt.neutron_sld(self.natural, wavelength=5)[0]
        self.H_sld = pt.neutron_sld(self.H, wavelength=5)[0]
        self.D_sld = pt.neutron_sld(self.D, wavelength=5)[0]

        self.D2Omatch = D2Omatch(self.H_sld, self.D_sld)
        self.charge = charge

    def D2Osld(self, volume_fraction=1.0, D2O_fraction=0.0):
        """
        Neutron SLD of the molecule in a %D2O solvent.
        """
        solvent_sld = D2O_fraction * D2O_SLD + (1 - D2O_fraction) * H2O_SLD
        solute_sld = D2O_fraction * self.D_sld + (1 - D2O_fraction) * self.H_sld
        return volume_fraction * solute_sld + (1 - volume_fraction) * solvent_sld


class Sequence(Molecule):
    """
    Convert FASTA sequence into chemical formula.

    *name* sequence name
    *sequence* code string
    *type* aa|dna|rna
       aa: amino acid sequence
       dna: dna sequence
       rna: rna sequence

    Note: rna sequence files treat T as U and dna sequence files treat U as T.
    """

    @staticmethod
    def loadall(filename):
        """
        Iterate over sequences in FASTA file, loading each in turn.

        Yields one FASTA sequence each cycle.
        """
        with open(filename, "rt") as fh:
            for name, seq in read_fasta(fh):
                yield Sequence(name, seq)

    @staticmethod
    def load(filename):
        """
        Load the first FASTA sequence from a file.
        """
        with open(filename, "rt") as fh:
            name, seq = next(read_fasta(fh))
            return Sequence(name, seq)

    def __init__(self, name, sequence, type="aa"):
        codes = _CODE_TABLES[type]
        parts = tuple(codes[c] for c in sequence)
        cell_volume = sum(p.cell_volume for p in parts)
        charge = sum(p.charge for p in parts)
        structure = []
        for p in parts:
            structure.extend(list(p.formula.structure))
        formula = pt.formula(structure).hill

        Molecule.__init__(self, name, formula, cell_volume=cell_volume, charge=charge)
        self.sequence = sequence


# Water density at 20C; neutron wavelength doesn't matter (use 5 A).
H2O_SLD = pt.neutron_sld(pt.formula("H2O@0.9982"), wavelength=5)[0]
D2O_SLD = pt.neutron_sld(pt.formula("D2O@0.9982"), wavelength=5)[0]


def D2Omatch(H_sld, D_sld):
    """
    Find the D2O% concentration of solvent such that neutron SLD of the
    material matches the neutron SLD of the solvent.

    *H_sld*, *D_sld* are the SLDs for the hydrogenated and deuterated forms
    of the material respectively, where *D* includes all the labile protons
    swapped for deuterons.  Water SLD is calculated at 20 C.

    Note that the resulting percentage is only meaningful between
    0% to 100%.  Beyond 100% you will need an additional constrast agent
    in the 100% D2O solvent to increase the SLD enough to match.
    """
    # SLD(%Dsample + (1-%)Hsample) = SLD(%D2O + (1-%)H2O)
    # %SLD(Dsample) + (1-%)SLD(Hsample) = %SLD(D2O) + (1-%)SLD(H2O)
    # %(SLD(Dsample) - SLD(Hsample) + SLD(H2O) - SLD(D2O)) = SLD(H2O) - SLD(Hsample)
    # % = 100*(SLD(H2O) - SLD(Hsample)) / (SLD(Dsample) - SLD(Hsample) + SLD(H2O) - SLD(D2O))
    return 100 * (H2O_SLD - H_sld) / (D_sld - H_sld + H2O_SLD - D2O_SLD)


def read_fasta(fp):
    """
    Iterate over the sequences in a FASTA file.

    Each iteration is a pair (sequence name, sequence codes).
    """
    name, seq = None, []
    for line in fp:
        line = line.rstrip()
        if line.startswith(">"):
            if name:
                yield (name, "".join(seq))
            name, seq = line, []
        else:
            seq.append(line)
    if name:
        yield (name, "".join(seq))


def isotope_substitution(formula, source, target, portion=1):
    """
    Substitute one atom/isotope in a formula with another in some proportion.

    *formula* is the formula being updated.

    *source* is the isotope/element to be substituted.

    *target* is the replacement isotope/element.

    *portion* is the proportion of source which is substituted for target.
    """
    # Note: formula.atoms creates a new structure each time
    atoms = formula.atoms
    if source in atoms:
        mass = formula.mass
        mass_reduction = atoms[source] * portion * (source.mass - target.mass)
        density = formula.density * (mass - mass_reduction) / mass
        atoms[target] = atoms.get(target, 0) + atoms[source] * portion
        if portion == 1:
            del atoms[source]
        else:
            atoms[source] *= 1 - portion
    else:
        density = formula.density
    return pt.formula(atoms, density=density)


# FASTA code table
def _(code, V, formula, name):
    if formula[-1] == "-":
        charge = -1
        formula = formula[:-1]
    elif formula[-1] == "+":
        charge = +1
        formula = formula[:-1]
    else:
        charge = 0
    molecule = Molecule(name, formula, cell_volume=V, charge=charge)
    molecule.code = code  # Add code attribute so we can write as well as read
    return code, molecule


AMINO_ACID_CODES = dict(
    (
        # code, volume, formula,        name
        _("A", 91.5, "C3H4TNO", "alanine"),
        _("C", 105.6, "C3H3TNOS", "cysteine"),
        _("D", 124.5, "C4H3TNO3-", "aspartic acid"),
        _("E", 155.1, "C5H5TNO3-", "glutamic acid"),
        _("F", 203.4, "C9H8TNO", "phenylalanine"),
        _("G", 66.4, "C2H2TNO", "glycine"),
        _("H", 167.3, "C6H5T3N3O+", "histidine"),
        _("I", 168.8, "C6H10TNO", "isoleucine"),
        _("K", 171.3, "C6H9T4N2O+", "lysine"),
        _("L", 168.8, "C6H10TNO", "leucine"),
        _("M", 170.8, "C5H8TNOS", "methionine"),
        _("N", 135.2, "C4H3T3N2O2", "asparagine"),
        _("P", 129.3, "C5H7NO", "proline"),
        _("Q", 161.1, "C5H5T3N2O2", "glutamine"),
        _("R", 202.1, "C6H7T6N4O+", "arginine"),
        _("S", 99.1, "C3H3T2NO2", "serine"),
        _("T", 122.1, "C4H5T2NO2", "threonine"),
        _("V", 141.7, "C5H8TNO", "valine"),
        _("W", 237.6, "C11H8T2N2O", "tryptophan"),
        _("Y", 203.6, "C9H7T2NO2", "tyrosine"),
    )
)


def _(formula, V, name):
    molecule = Molecule(name, formula, cell_volume=V)
    return name, molecule


NUCLEIC_ACID_COMPONENTS = dict(
    (
        # formula, volume, name
        _("NaPO3", 60, "phosphate"),
        _("C5H6TO3", 125, "ribose"),
        _("C5H7O2", 115, "deoxyribose"),
        _("C5H2T2N5", 114, "adenine"),
        _("C4H2TN2O2", 99, "uracil"),
        _("C5H4TN2O2", 126, "thymine"),
        _("C5HT3N5O", 119, "guanine"),
        _("C4H2T2N3O", 103, "cytosine"),
    )
)

CARBOHYDRATE_RESIDUES = dict(
    (
        # formula, volume, name
        _("C6H7T3O5", 171.9, "Glc"),
        _("C6H7T3O5", 166.8, "Gal"),
        _("C6H7T3O5", 170.8, "Man"),
        _("C6H7T4O5", 170.8, "Man (terminal)"),
        _("C8H10T3NO5", 222.0, "GlcNAc"),
        _("C8H10T3NO5", 232.9, "GalNAc"),
        _("C6H7T3O4", 160.8, "Fuc (terminal)"),
        _("C11H11T5NO8", 326.3, "NeuNac (terminal)"),
        # Glycosaminoglycans
        _("C14H15T5NO11Na", 390.7, "hyaluronate"),  # GlcA.GlcNAc
        _("C14H17T5NO13SNa", 473.5, "keratan sulphate"),  # Gal.GlcNAc.SO4
        _("C14H15T4NO14SNa", 443.5, "chondroitin sulphate"),  # GlcA.GalNAc.SO4
    )
)

LIPIDS = dict(
    (
        # formula, volume, name
        _("CH2", 27, "methylene"),
        _("CD2", 27, "methylene-D"),
        _("C10H18NO8P", 350, "phospholipid headgroup"),
        _("C6H5O6", 240, "triglyceride headgroup"),
        _("C36H72NO8P", 1089, "DMPC"),
        _("C36H20D52NO8P", 1089, "DMPC-D52"),
        _("C29H55T3NO8P", 932, "DLPE"),
        _("C27H45TO", 636, "cholesteral"),
        _("C45H78O2", 1168, "oleate"),
        _("C57H104O6", 1617, "trioleate form"),
        _("C39H77T2N2O2P", 1166, "palmitate ester"),
    )
)


def _(code, formula, V, name):
    molecule = Molecule(name, formula, cell_volume=V)
    molecule.code = code
    return code, molecule


RNA_BASES = dict(
    (
        # code, formula, volume, name
        _("A", "C10H8T3N5O6PNa", 299, "adenosine"),
        _("T", "C9H8T2N2O8PNa", 284, "uridine"),  # Use T for U in RNA
        _("G", "C10H7T4N5O7PNa", 304, "guanosine"),
        _("C", "C9H8T3N3O7PNa", 288, "cytidine"),
    )
)

DNA_BASES = dict(
    (
        # code, formula, volume, %D2O matchpoint, name
        _("A", "C10H9T2N5O5PNa", 289, "adenosine"),
        _("T", "C10H11T1N2O7PNa", 301, "thymidine"),
        _("G", "C10H8T3N5O6PNa", 294, "guanosine"),
        _("C", "C9H9T2N3O6PNa", 278, "cytidine"),
    )
)


def _nucleic_acid_average(bases, code_table):
    """
    Compute average over possible nucleotides, assuming equal weight if
    precise nucleotide is not known
    """
    n = len(bases)
    D, cell_volume = pt.formula(), 0
    for c in bases:
        D += code_table[c].formula
        cell_volume += code_table[c].cell_volume
    if n > 0:
        D, cell_volume = (1 / n) * D, cell_volume / n
    return D, cell_volume


def _(code, bases, name):
    D, V = _nucleic_acid_average(bases, RNA_BASES)
    rna = Molecule(name, D.hill, cell_volume=V)
    rna.code = code
    D, V = _nucleic_acid_average(bases, DNA_BASES)
    dna = Molecule(name, D.hill, cell_volume=V)
    dna.code = code
    return (code, rna), (code, dna)


RNA_CODES, DNA_CODES = [
    dict(v)
    for v in zip(
        # code, nucleotides,  name
        _("A", "A", "adenosine"),
        _("C", "C", "cytidine"),
        _("G", "G", "guanosine"),
        _("T", "T", "thymidine"),
        _("U", "T", "uridine"),  # RNA_BASES["T"] is uridine
        _("R", "AG", "purine"),
        _("Y", "CT", "pyrimidine"),
        _("K", "GT", "ketone"),
        _("M", "AC", "amino"),
        _("S", "CG", "strong"),
        _("W", "AT", "weak"),
        _("B", "CGT", "not A"),
        _("D", "AGT", "not C"),
        _("H", "ACT", "not G"),
        _("V", "ACG", "not T"),
        _("N", "ACGT", "any base"),
        _("X", "", "masked"),
        _("-", "", "gap"),
    )
]


_CODE_TABLES = {
    "aa": AMINO_ACID_CODES,
    "dna": DNA_CODES,
    "rna": RNA_CODES,
}


def fasta_table():
    rows = []
    rows += [v for k, v in sorted(AMINO_ACID_CODES.items())]
    rows += [v for k, v in sorted(NUCLEIC_ACID_COMPONENTS.items())]
    rows += [Sequence("beta casein", beta_casein)]

    print(
        "%20s %7s %7s %7s %5s %5s %5s %5s %5s %5s"
        % ("name", "M(H2O)", "M(D2O)", "volume", "den", "#el", "xray", "nH2O", "nD2O", "%D2O match")
    )
    for v in rows:
        protons = sum(num * el.number for el, num in v.formula.atoms.items())
        electrons = protons - v.charge
        xray_sld = pt.xray_sld(v.formula, wavelength=pt.Cu.K_alpha)
        print(
            "%20s %7.1f %7.1f %7.1f %5.2f %5d %5.2f %5.2f %5.2f %5.1f"
            % (
                v.name,
                v.H.mass,
                v.D.mass,
                v.cell_volume,
                v.natural.density,
                electrons,
                xray_sld[0],
                v.H_sld,
                v.D_sld,
                v.D2Omatch,
            )
        )


beta_casein = "RELEELNVPGEIVESLSSSEESITRINKKIEKFQSEEQQQTEDELQDKIHPFAQTQSLVYPFPGPIPNSLPQNIPPLTQTPVVVPPFLQPEVMGVSKVKEAMAPKHKEMPFPKYPVEPFTESQSLTLTDVENLHLPLPLLQSWMHQPHQPLPPTVMFPPQSVLSLSQSKVLPVPQKAVPYPQRDMPIQAFLLYQEPVLGPVRGPFPIIV"


def test():
    from periodictable.constants import avogadro_number

    # Beta casein results checked against Duncan McGillivray's spreadsheet
    # beta casein 23561.9 23880.9 30872.9  1.27 12614 11.55  1.68  2.75
    s = Sequence("beta casein", beta_casein)
    assert abs(s.D.mass - 23880.9) < 0.1
    # print "density", s.mass/avogadro_number/s.cell_volume*1e24
    assert abs(s.natural.mass / avogadro_number / s.cell_volume * 1e24 - 1.267) < 0.01
    assert abs(s.D_sld - 2.75) < 0.01

    # Check that X-ray sld is independent of isotope
    H = isotope_substitution(s.formula, pt.T, pt.H)
    D = isotope_substitution(s.formula, pt.T, pt.D)
    H_sld, D_sld = pt.xray_sld(H, wavelength=1.54), pt.xray_sld(D, wavelength=1.54)
    # print H_sld, D_sld
    assert abs(H_sld[0] - D_sld[0]) < 1e-10


def main():
    import sys
    import os

    if len(sys.argv) == 1 or sys.argv[1] in ("-h", "--help", "-?"):
        print("usage: python -m periodictable.fasta file|sequence [%D]")
        sys.exit(1)
    if os.path.exists(sys.argv[1]):
        seq_list = Sequence.loadall(sys.argv[1])
    else:
        seq_list = [Sequence("", sys.argv[1])]
    if len(sys.argv) > 2:
        concentration = 0.01 * float(sys.argv[2])
    else:
        concentration = 0.50

    for seq in seq_list:
        mixture_sld = seq.H_sld * (1 - concentration) + seq.D_sld * concentration
        print(
            "%s mass:%.1f H-sld: %.3f  D-sld: %.3f  D[%g%%]-sld: %.3f"
            % (seq.name, seq.natural.mass, seq.H_sld, seq.D_sld, 100 * concentration, mixture_sld)
        )


if __name__ == "__main__":
    # main()
    fasta_table()



import numba
import numpy as np
from math import erf

from .clone_module import clone_module

MODULE = clone_module("refl1d.lib.python.build_profile")

SQRT1_2 = 1.0 / np.sqrt(2.0)


@numba.vectorize
def verf(x):
    return erf(x)


@numba.njit(cache=True)
def blend(z, sigma, offset):
    """
    blend function

    Given a Gaussian roughness value, compute the portion of the neighboring
    profile you expect to find in the current profile at depth z.
    """
    if sigma <= 0.0:
        return 1.0 * (z >= offset)
    else:
        return 0.5 * verf(SQRT1_2 * (z - offset) / sigma) + 0.5


MODULE.blend = blend

# BUILD_PROFILE_SIG = 'void(f8[:], f8[:], f8[:], f8[:], f8[:], f8[:,:])'
# build_profile = numba.njit(BUILD_PROFILE_SIG)(MODULE.build_profile)
build_profile = numba.njit(cache=True)(MODULE.build_profile)
MODULE.build_profile = build_profile



# This program is in the public domain
# Author: Paul Kienzle
"""
Refl1D: Specular Reflectometry Modeling and Fitting Software

This package provides tools for modeling a variety of systems from
simple slabs to complex systems with a mixture models involving
smoothly varying layers.

X-ray and neutron and polarized neutron data can be loaded and the model
parameters adjusted to achieve the best fit.

A graphical interface allows direct manipulation of the model profiles.

See http://refl1d.readthedocs.org for online manuals.
"""

import os
import sys
from typing import Literal

try:
    from ._version import __version__  # noqa: F401
except ImportError:
    __version__ = "unknown"


BACKEND_NAMES = Literal["numba", "c_ext", "python"]
BACKEND_NAME: BACKEND_NAMES = os.environ.get("REFL1D_BACKEND", "numba")


def use(backend_name: BACKEND_NAMES):
    global BACKEND_NAME
    BACKEND_NAME = backend_name
    if "refl1d.refllib" in sys.modules:
        # then it's already been imported:
        import refl1d.refllib

        refl1d.refllib.set_backend(backend_name)



import numba
from .clone_module import clone_module

MODULE = clone_module("refl1d.lib.python.convolve")

MODULE.prange = numba.prange

convolve_uniform = numba.njit("(f8[:], f8[:], f8[:], f8[:], f8[:])", cache=True, parallel=False)(
    MODULE.convolve_uniform
)

convolve_gaussian_point = numba.njit(
    "f8(f8[:], f8[:], i8, i8, f8, f8, f8)",
    cache=True,
    parallel=False,
    locals={
        "z": numba.float64,
        "Glo": numba.float64,
        "erflo": numba.float64,
        "erfmin": numba.float64,
        "y": numba.float64,
        "zhi": numba.float64,
        "Ghi": numba.float64,
        "erfhi": numba.float64,
        "m": numba.float64,
        "b": numba.float64,
    },
)(MODULE.convolve_gaussian_point)
MODULE.convolve_gaussian_point = convolve_gaussian_point

# has same performance when using guvectorize instead of njit:
# @numba.guvectorize("(i8, f8[:], f8[:], i8, f8[:], f8[:], f8[:])", '(),(m),(m),(),(n),(n)->(n)')

convolve_gaussian = numba.njit(
    "(f8[:], f8[:], f8[:], f8[:], f8[:])",
    cache=True,
    parallel=False,
    locals={
        "sigma": numba.float64,
        "xo": numba.float64,
        "limit": numba.float64,
        "k_in": numba.int64,
        "k_out": numba.int64,
    },
)(MODULE.convolve_gaussian)



"""
Data stitching.

Join together datasets yielding unique sorted x.
"""

from numpy import hstack, vstack, argsort, sum, sqrt


def stitch(data, same_x=0.001, same_dx=0.001):
    """
    Stitch together multiple measurements into one.

    *data* a list of datasets with x, dx, y, dy attributes
    *same_x* minimum point separation (default is 0.001).
    *same_dx* minimum change in resolution that may be averaged (default is 0.001).

    WARNING: the returned x values may be data dependent, with two measured
    sets having different x after stitching, even though the measurement
    conditions are identical!!

    Either add an intensity weight to the datasets::

        probe.I = slitscan

    or use interpolation if you need to align two stitched scans::

        import numpy as np
        x1, dx1, y1, dy1 = stitch([a1, b1, c1, d1])
        x2, dx2, y2, dy2 = stitch([a2, b2, c2, d2])
        x2[0], x2[-1] = x1[0], x1[-1] # Force matching end points
        y2 = np.interp(x1, x2, y2)
        dy2 = np.interp(x1, x2, dy2)
        x2 = x1

    WARNING: the returned dx value underestimates the true x, depending on
    the relative weights of the averaged data points.
    """
    if same_dx is None:
        same_dx = same_x
    x = hstack(p.x for p in data)
    dx = hstack(p.dx for p in data)
    y = hstack(p.y for p in data)
    dy = hstack(p.dy for p in data)
    if all(hasattr(p, "I") for p in data):
        weight = hstack(p.I for p in data)
    else:
        weight = y / dy**2  # y/dy**2 is approximately the intensity

    # Sort the data by increasing x
    idx = argsort(x)
    data = vstack((x, dx, y, dy, weight))
    data = data[:, idx]
    x = data[0, :]
    dx = data[1, :]

    # Skip through the data looking for regions of overlap.
    keep = []
    n, last, next = len(x), 0, 1
    while next < n:
        while next < n and abs(x[next] - x[last]) <= same_x:
            next += 1
        if next - last == 1:
            # Only one point, so no averaging necessary
            keep.append(last)
        else:
            # Pick the x in [last:next] with the best resolution and average
            # them using Poisson averages.  Repeat until all points are used
            remainder = data[:, last:next]
            avg = []
            while remainder.shape[1] > 0:
                best_dx = min(remainder[1, :])
                idx = remainder[1, :] - best_dx <= same_dx
                avg.append(poisson_average(remainder[:, idx]))
                remainder = remainder[:, ~idx]
            # Store the result in worst to best resolution order.
            for i, d in enumerate(reversed(avg)):
                data[:, last + i] = d
                keep.append(last + i)
        last = next

    return data[:4, keep]


def poisson_average(xdxydyw):
    """
    Compute the poisson average of y/dy using a set of data points.

    The returned x, dx is the weighted average of the inputs::

        x = sum(x*I)/sum(I)
        dx = sum(dx*I)/sum(I)

    The returned y, dy use Poisson averaging::

        w = sum(y/dy^2)
        y = sum((y/dy)^2)/w
        dy = sqrt(y/w)

    The above formula gives the expected result for combining two
    measurements, assuming there is no uncertainty in the monitor.

    ::

        measure N counts during M monitors
          rate:                   r = N/M
          rate uncertainty:       dr = sqrt(N)/M
          weighted rate:          r/dr^2 = (N/M) / (N/M^2) =  M
          weighted rate squared:  r^2/dr^2 = (N^2/M^2) / (N/M^2) = N

        for two measurements Na, Nb
          w = ra/dra^2 + rb/drb^2 = Ma + Mb
          y = ((ra/dra)^2 + (rb/drb)^2)/w = (Na + Nb)/(Ma + Mb)
          dy = sqrt(y/w) = sqrt( (Na + Nb)/ w^2 ) = sqrt(Na+Nb)/(Ma + Mb)

    This formula isn't strictly correct when applied to values which
    have been scaled, for example to account for an attenuator in the
    counting system.
    """
    # TODO: need better estimate of dx, with weighted broadening according
    # to the distance of the x's from the centers.
    # TODO: check the accuracy of the formula in the presence of
    # attenuators.
    x, dx, y, _dy, weight = xdxydyw
    w = sum(weight)
    x = sum(x * weight) / w
    dx = sum(dx * weight) / w
    y = sum(y * weight) / w
    dy = sqrt(y / w)
    # print "averaging", xdxydy, x, dx, y, dy
    return x, dx, y, dy, w



# This program is public domain.
r"""
Optical matrix form of the reflectivity calculation.

This is a pure python implementation of reflectometry which returns the
reflection and transmission at each layer.  Based on C-D matrix form
given in [1].

***WARNING*** Returns the conjugate of the reflectometry amplitude compared
to the abeles matrix calculation given on wikipedia.[2]

***WARNING*** Calculation not yet trusted: $r^2 + t^2 \ne 1$ for perfect
interfaces with no absorption.  However, $r^2 + t^2 k0/kn = 1$.  This detail
comes from the tmm package transmission power function.[3]

***WARNING*** Transmission value explodes below the critical edge.

***WARNING*** Nevot-Croce does not preserve $r^2 + t^2 k0/kn = 1$

Note: for conversion to GPU with parallel prefix matrix multiplication[4],
need to restructure the calculation so the matrices are independent, and
we don't need to first perform the cumulative sum on depth z.  The Abeles
matrix algorithm[2] does this, but it is not computing transmissions
correctly.

[1] Dura, J.A., Rus, E.D., Kienzle, P.A., Maranville, B.B., 2017.
Nanolayer Analysis by Neutron Reflectometry,
in: Imae, T. (Ed.), Nanolayer Research. Elsevier, Amsterdam, pp. 155-202.
https://doi.org/10.1016/B978-0-444-63739-0.00005-0

[2] https://en.wikipedia.org/wiki/Transfer-matrix_method_(optics)#Abeles_matrix_formalism

[3] https://github.com/sbyrnes321/tmm/blob/4ebbe9dbee140ccabad22c158107681846c9d204/tmm_core.py#L161-L181

[4] https://en.wikipedia.org/wiki/Prefix_sum
"""

from __future__ import print_function, division

import numpy as np
from numpy import asarray, isscalar, empty, ones, ones_like
from numpy import sqrt, exp, pi


def refl_tr(kz, depth, rho, irho=0, sigma=0, rho_index=None):
    r"""
    Reflectometry as a function of kz for a set of slabs.

    :Parameters:

    *kz* : float[n] | |1/Ang|
        Scattering vector $2\pi\sin(\theta)/\lambda$. This is $\tfrac12 Q_z$.
    *depth* :  float[m] | |Ang|
        thickness of each layer.  The thickness of the incident medium
        and substrate are ignored.
    *rho*, *irho* :  float[n, k] | |1e-6/Ang^2|
        real and imaginary scattering length density for each layer for each kz
        Note: absorption cross section mu = 2 irho/lambda
    *sigma* : float[m-1] | |Ang|
        interfacial roughness.  This is the roughness between a layer
        and the subsequent layer.  There is no interface associated
        with the substrate.  The sigma array should have at least m-1
        entries, though it may have m with the last entry ignored.
    *rho_index* : int[m]
        index into rho vector for each kz

    Returns array of shape [k X 2 X n], with R[k, 0, n] as the transmission
    amplitude and R[k, 1, n] as the reflection amplitude at the layer k
    interface for kz[n].

    Slabs are ordered with the incident SLD at index 0 and backing at
    index -1.
    """
    if isscalar(kz):
        kz = asarray([kz], "d")

    # Don't support kz negative for now
    if (kz < 0).any():
        raise ValueError("require kz >= 0")

    m = len(depth)

    # Make everything into arrays
    depth = asarray(depth, "d")
    rho = asarray(rho, "d")
    irho = irho * ones_like(rho) if isscalar(irho) else asarray(irho, "d")
    sigma = sigma * ones(m - 1, "d") if isscalar(sigma) else asarray(sigma, "d")

    # Repeat rho, irho columns as needed
    if rho_index is not None:
        rho = rho[rho_index, :]
        irho = irho[rho_index, :]
    elif len(rho.shape) == 1:
        rho = rho[None, :]
        irho = irho[None, :]

    # Force the correct branch cut for sqrt below the critical edge
    irho = abs(irho) + 1e-30

    r = _calc(kz, depth, rho, irho, sigma)
    return r


def _calc(kz, depth, rho, irho, sigma):
    # type: (np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray) -> np.ndarray

    num_kz = len(kz)
    num_interfaces = len(depth) - 1

    # Complex index of refraction is relative to the incident medium.
    # Dura, J.A., Rus, E.D., Kienzle, P.A., Maranville, B.B., 2017.
    # Nanolayer Analysis by Neutron Reflectometry,
    # in: Imae, T. (Ed.), Nanolayer Research. Elsevier, Amsterdam, pp. 155-202.
    # https://doi.org/10.1016/B978-0-444-63739-0.00005-0
    #
    # Neutron energy in the layer, E0, is:
    #     E0 = h_bar^2/(2 m_n) (k_l^2 + 4 pi rho_l)
    # For the incident neutron, E0 is defined by k_z and incident rho:
    #     E0 = h_bar^2/(2 m_n) (k_z^2 + 4 pi rho_in)
    # Solving for k_l:
    #     k_l = sqrt((2 m_n)/h_bar E0 - 4 pi rho_l)
    # Substituting E0:
    #     k_l = sqrt(k_z^2 + 4 pi rho_in - 4 pi rho_l)
    # rho_in = np.where(kz >= 0, rho[:, 0], rho[:, -1])
    rho_in = rho[:, 0]
    kz_sq = kz * kz

    k = sqrt(kz_sq + 4e-6 * pi * (rho_in - rho[:, 0]))
    z = 0
    M = []
    k_list = [k]
    B11 = 1
    B22 = 1
    B21 = 0
    B12 = 0
    for i in range(num_interfaces):
        k_next = sqrt(kz_sq + 4e-6 * pi * (rho_in - rho[:, i + 1] + 1j * irho[:, i + 1]))
        nevot_croce_damping = exp(-2 * k * k_next * sigma[i] ** 2)
        k_plus, k_minus = k_next + k, k_next - k
        M11 = k_plus * exp(-1j * k_minus * z) / (2 * k)
        M22 = k_plus * exp(+1j * k_minus * z) / (2 * k)
        M12 = nevot_croce_damping * k_minus * exp(-1j * k_plus * z) / (2 * k)
        M21 = nevot_croce_damping * k_minus * exp(+1j * k_plus * z) / (2 * k)
        M.append((M11, M12, M21, M22))
        k_list.append(k_next)
        # Right-multiply: B = M*B
        C11 = M11 * B11 + M12 * B21
        C21 = M21 * B11 + M22 * B21
        B11, B21 = C11, C21
        C12 = M11 * B12 + M12 * B22
        C22 = M21 * B12 + M22 * B22
        B12, B22 = C12, C22
        k = k_next
        z += depth[i + 1]
        # print("==== layer %d ====" % i)
        # print("B11:", B11)
        # print("B22:", B22)
        # print("B21:", B21)
        # print("B12:", B12)

    r = -B21 / B22
    t = B11 + B12 * r

    results = np.empty((num_interfaces + 1, 2, num_kz), "D")
    if 1:
        # Propagate (1, r) forward using:
        #    [c_{n+1}, d_{n+1}]^T = M_n [c_n, d_n]^T
        # print("propagate forward")
        c, d = np.ones(num_kz, "D"), r
        results[0] = c, d
        for i in range(num_interfaces):
            M11, M12, M21, M22 = M[i]
            c, d = M11 * c + M12 * d, M21 * c + M22 * d
            results[i + 1] = c, d
    else:
        # Propagate (t, 0) backward using:
        #    [c_n, d_n]^T = M_n^{-1} [c_{n+1}, d_{n+1}]^T
        # print("propagate backward")
        c, d = t, np.zeros(num_kz, "D")
        results[-1] = c, d
        for i in reversed(range(num_interfaces)):
            M11, M12, M21, M22 = M[i]
            # Minv = [[M22, -M12], [-M21, M11]] / (M11 M22 - M12 M21)
            det = B11 * B22 - B21 * B12
            c, d = (M22 * c - M12 * d) / det, (-M21 * c + M11 * d) / det
            results[i] = c, d

    # print("r   ", r)
    # for k, layer in enumerate(results):
    #    print("r[%d]"%k, layer[1])
    #    print("t[%d]"%k, layer[0])
    # print("t   ", t)
    k0, kn = k_list[0], k_list[-1]
    # print("r^2 =", abs(r**2))
    # print("t^2 =", abs(t**2))
    # print("r^2 + t^2*k0/kn =", abs(r**2) + abs(t**2)*(k0/kn).real)
    # print("k0", k0)
    # print("kn", kn)

    ## Compute back reflectivity
    # negative = (kz < 0.0)
    # if negative.any():
    #    det = B11*B22 - B21*B12
    #    r[negative] = -(B12 / B22 / det)[negative]
    return results


def check():
    import numpy as np

    np.set_printoptions(linewidth=10000)

    # q = np.linspace(-0.3, 0.3, 4)
    q = np.linspace(0.1, 0.3, 3)
    layers = [
        # depth rho irho sigma
        [0, 0.0, 0.0, 0.0],
        [200, 2.0, 0.0, 0.0],
        [200, 4.0, 0.0, 0.0],
        [0, 3.0, 0.0, 0.0],
    ]
    # layers[1][2] = 1.0  # add absorption to layer 1
    # for L in layers[:-1]: L[3] = 10.0   # add roughness to all layers

    depth, rho, irho, sigma = (np.asarray(v) for v in zip(*layers))
    rho *= 100  # show point below critical edge
    print("q", q)
    try:
        from .abeles import refl

        r_old = refl(q / 2, depth, rho, irho=irho, sigma=sigma)
        print("rold", r_old)
    except ImportError:
        print("could not import abeles")
    wave = refl_tr(q / 2, depth, rho, irho=irho, sigma=sigma)
    for k, wk in enumerate(wave):
        print("r[%d]" % k, wk[1])
        print("t[%d]" % k, wk[0])
    r, t = wk[0][1], wk[-1][0]
    # print("r^2", abs(r**2))


def check2(t=200, rho=4.66):
    import numpy as np

    np.set_printoptions(linewidth=10000)

    # q = np.linspace(0.0, 0.05, 600)
    q = np.linspace(0.0, 0.0005, 6000)
    # q = np.linspace(0.1, 0.3, 3)
    layers = [
        # depth rho irho sigma
        [0, 0.0, 0.0, 0.0],
        # [200, 2.0, 0.0,  0.0],
        [t, -rho, 0.0, 0.0],
        [0, 0, 0.0, 0.0],
        # [  0, 2.07, 0.0,  0.0],
    ]
    # add absorption
    # layers[1][2] = 1.0
    depth, rho, irho, sigma = zip(*layers)

    wave = refl_tr(q / 2, depth, rho, irho=irho, sigma=sigma)
    # print("shape", wave.shape)
    r = wave[0, 1]

    import matplotlib.pyplot as plt

    plt.plot(q, r.real, "-b", label="real")
    plt.plot(q, r.imag, "-r", label="imag")
    plt.plot(q, abs(r**2), "-g", label="magnitude")
    # plt.legend()
    plt.grid(True)
    # plt.xlabel('q (1/A)')
    # plt.title("stack = Si / 20 nm Au / Air")
    # plt.show()
    # print("q", q)
    # print("r", r)
    # print("r^2", abs(r**2))


def check3(t=200, rho=4.66):
    import numpy as np

    np.set_printoptions(linewidth=10000)

    qmax = 0.1
    t, qmax = 2000, 0.05
    # t, qmax = 20, 0.4
    q = np.linspace(0.0, qmax, 6000)[1:]
    layers = [
        # depth rho irho sigma
        [0, 0.0, 0.0, 0.0],
        # [200, 2.0, 0.0,  0.0],
        [t, rho, 0.0, 0.0],
        [0, 0, 0.0, 0.0],
        # [  0, 2.07, 0.0,  0.0],
    ]
    # add absorption
    # layers[1][2] = 1.0
    vdepth, vrho, virho, vsigma = [np.asarray(v) for v in zip(*layers)]

    if 1:
        steps = 2001
        # portion = 0.9
        portion = 0.0
        thickness = t
        flat = thickness * portion
        ends = thickness * (1 - portion)
        z = np.linspace(-3, 3, steps)
        vrho = np.exp(-0.5 * z**2) * rho
        virho = 0 * rho
        vdepth = np.ones_like(z) * ends / steps
        vdepth[int((steps - 1) / 2)] = flat
        vsigma = 0 * rho

    if 0:
        steps = 2001
        thickness, portion = t, 0.0
        flat = thickness * portion
        ends = thickness * (1 - portion)
        z = np.linspace(-3, 3, steps)
        vrho = np.exp(-0.5 * z**2) * rho
        vrho += -np.exp(-0.5 * (z - 1) ** 2 * 4) * rho
        virho = 0 * rho
        vdepth = np.ones_like(z) * ends / steps
        vdepth[int((steps - 1) / 2)] = flat
        vsigma = 0 * rho

    ## normalize to same amount of scattering density
    # total = rho*t
    # vrho *= total/np.sum(vrho*vdepth)

    wave = refl_tr(q / 2, vdepth, vrho, irho=virho, sigma=vsigma)
    # print("shape", layers.shape)
    r = wave[0, 1]

    import matplotlib.pyplot as plt
    from mpl_toolkits.axes_grid.inset_locator import InsetPosition

    scale = q * 0 + 1
    # scale = q**4
    plt.plot(q, scale * r.real, "-b", label="real", zorder=2)
    plt.plot(q, scale * r.imag, "-r", label="imag")
    plt.plot(q, scale * abs(r**2), "-g", label="magnitude")
    # plt.yscale('symlog', linthreshy=1e-6)
    # plt.legend()
    plt.grid(True)
    ax = plt.gca()
    inset = plt.axes([0, 0, 1, 1])
    inset.step(np.cumsum(vdepth), vrho, "-g", label="rho")
    # inset.step(np.cumsum(vdepth), virho, '-m', label='irho')
    inset.set_axes_locator(InsetPosition(ax, [0.6, 0.7, 0.4, 0.3]))
    inset.patch.set_alpha(0.5)
    inset.patch.set_color([0.9, 0.9, 0.8])
    # mark_inset(ax, inset)

    # plt.xlabel('q (1/A)')
    # plt.title("stack = Si / 20 nm Au / Air")
    # plt.show()
    # print("q", q)
    # print("r", r)
    # print("r^2", abs(r**2))


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    check3()
    """
    plt.subplot(211)
    check2(t=50)
    check2(t=100)
    check2(t=200)
    check2(t=400)
    check2(t=800)
    check2(t=1600)
    plt.subplot(212)
    check2(t=200,rho=-2)
    check2(t=200,rho=1)
    check2(t=200,rho=2)
    check2(t=200,rho=4)
    check2(t=200,rho=6)
    check2(t=200,rho=8)
    """
    plt.show()



import numba
from .clone_module import clone_module

MODULE = clone_module("refl1d.lib.python.reflectivity")

_REFL_SIG = "c16(i8, f8, f8[:], f8[:], f8[:], f8[:])"
_REFL_LOCALS = {
    "cutoff": numba.float64,
    "next": numba.int64,
    "sigma_offset": numba.int64,
    "step": numba.int8,
    "pi4": numba.float64,
    "kz_sq": numba.float64,
    "k": numba.complex128,
    "k_next": numba.complex128,
    "F": numba.complex128,
    "J": numba.complex128,
}
_REFL_LOCALS.update(("B{i}{j}".format(i=i, j=j), numba.complex128) for i in range(1, 3) for j in range(1, 3))
_REFL_LOCALS.update(("M{i}{j}".format(i=i, j=j), numba.complex128) for i in range(1, 3) for j in range(1, 3))
_REFL_LOCALS.update(("C{i}".format(i=i), numba.complex128) for i in range(1, 3))

refl = numba.njit(_REFL_SIG, parallel=False, cache=True, locals=_REFL_LOCALS)(MODULE.refl)
MODULE.refl = refl

REFLAMP_SIG = "void(f8[:], f8[:], f8[:,:], f8[:,:], f8[:], i4[:], c16[:])"

reflectivity_amplitude = numba.njit(REFLAMP_SIG, parallel=False, cache=True, locals={"offset": numba.int64})(
    MODULE.reflectivity_amplitude
)



"""
Residual correlation test.

INCOMPLETE UNUSED UNTESTED CODE

Returns the probability of seeing the given structure in the residuals.

The residuals are assumed to be e ~ N(0, 1).  Gross deviations from this
form will already be accounted for in the chisq statistic, and can be
ignored here.  We do this by translating the residual into a z-scored
residual, which should be in

Protects against an incorrect estimation of the uncertainty in the data,
which is anyway already covered by the chisq cost function.
"""

from __future__ import print_function

from scipy import stats
import numpy as np


def residual_nllf(v):
    # Normalize the scores so that we are assuming
    z = stats.zscore(v)


def plot(self):
    import matplotlib.pyplot as plt

    plt.subplot(313)
    Q, R = self.reflectivity()
    self.probe.plot_transform(theory=(Q, R), substrate=self.sample[0].material, surround=self.sample[-1].material)


# Experiments with different sorts of
def plot_transform(self, theory=None, substrate=None, surround=None):
    """
    Plot the Fresnel-normalized reflectivity associated with the probe.
    """
    import matplotlib.pyplot as plt

    if substrate is None and surround is None:
        raise TypeError("Fresnel-normalized reflectivity needs substrate or surround")
    F = self.fresnel(substrate=substrate, surround=surround)
    print("substrate %s" % substrate)
    # Qc = sqrt(16*pi*substrate)
    Qc = 0
    T = np.linspace(Qc, max(self.Q), len(self.Q))
    if hasattr(self, "R"):
        A = abs(np.fft.fft(np.interp(T, self.Q, self.R / F)))
        plt.plot(T, A, ".")
    if theory is not None:
        Q, R = theory
        A = abs(np.fft.fft(np.interp(T, self.Q, self.R / F)))
        plt.plot(T, A)
    plt.xlabel("z (Angstroms)")
    if substrate is None:
        name = "air:%s" % (surround.name)
    elif surround is None or isinstance(surround, Vacuum):
        name = substrate.name
    else:
        name = "%s:%s" % (substrate.name, surround.name)
    plt.ylabel("FFT(R/R(%s))" % (name))


def plot_deriv(self, theory=None):
    import matplotlib.pyplot as plt

    scale = 1e8 * self.Q**4
    # scale=1
    if hasattr(self, "R"):
        d = deriv(self.Q, self.R, self.dR)
        plt.plot(self.Q, d * scale, ".")
    if theory is not None:
        Q, R = theory
        dth = deriv(Q, R)
        plt.plot(Q, dth * scale)
    # plt.plot(Q, (sign(d)-sign(dth))/2)
    plt.xlabel("Q (inv Angstroms)")
    plt.ylabel("R' (100 Q)^4")


def deriv(Q, R, dR=None, width=5, degree=2):
    from bumps.wsolve import wpolyfit

    d = np.empty_like(Q)
    if dR is None:
        dR = np.ones_like(Q)
    k = (width - 1) / 2
    p = wpolyfit(Q[:width], R[:width], dy=dR[:width], degree=degree)
    d[: k + 1] = p.der(Q[: k + 1])
    for i in range(k + 1, len(Q) - k - 1):
        idx = slice(i - k, i + k + 1)
        p = wpolyfit(Q[idx], R[idx], dy=dR[idx], degree=degree)
        d[i] = p.der(Q[i])
    p = wpolyfit(Q[-width:], R[-width:], dy=dR[-width:], degree=degree)
    d[-k - 1 :] = p.der(Q[-k - 1 :])
    return d



"""
Exported names

In model definition scripts, rather than importing symbols one by one, you
can simply perform:

    from refl1d.names import *

This is bad style for library and applications but convenient for
small scripts.
"""

import sys
import numpy as np

from periodictable import elements, formula
from bumps.parameter import Parameter, FreeVariables
from bumps import pmath
from bumps.pdfwrapper import PDF
from bumps.fitproblem import FitProblem, MultiFitProblem
# from bumps.fitproblem import FitProblem  # deprecated

from .experiment import Experiment, plot_sample, MixedExperiment
from .flayer import FunctionalProfile, FunctionalMagnetism
from .material import SLD, Material, Compound, Mixture
from .model import Slab, Stack
from .polymer import PolymerBrush, PolymerMushroom, EndTetheredPolymer, VolumeProfile, layer_thickness
from .mono import FreeLayer, FreeInterface
from .cheby import FreeformCheby, ChebyVF, cheby_approx, cheby_points
from .probe import Probe, ProbeSet, XrayProbe, NeutronProbe, QProbe, PolarizedNeutronProbe, PolarizedQProbe, load4
from .stajconvert import load_mlayer, save_mlayer
from . import ncnrdata as NCNR, snsdata as SNS
from .instrument import Monochromatic, Pulsed
from .magnetic import MagneticSlab, MagneticTwist, FreeMagnetic, MagneticStack
from .magnetism import Magnetism, MagnetismTwist, FreeMagnetism, MagnetismStack
from .support import sample_data

# Pull in common materials for reflectometry experiments.
# This could lead to a lot of namespace pollution, and particularly to
# confusion if the user also does "from periodictable import *" since
# both of them create elements.
# Python doesn't allow "from .module import *"
from .materialdb import *


# Deprecated names
def ModelFunction(*args, **kw):
    raise NotImplementedError("ModelFunction no longer supported --- use PDF instead")


PolarizedNeutronQProbe = PolarizedQProbe
numpy = np



# This code is public domain

"""
Pure python Fresnel reflectivity calculator.
"""

from numpy import sqrt, exp, real, conj, pi, abs, choose


class Fresnel(object):
    """
    Function for computing the Fresnel reflectivity for a single interface.

    :Parameters:
        *rho*, *irho* = 0 : float | 1e6 * inv Angstrom^2
            real and imaginary scattering length density of backing medium
        *Vrho*, *Virho* = 0 : float | 1e6 * inv Angstrom^2
            real and imaginary scattering length density of incident medium
        *sigma* = 0 : float | Angstrom
            interfacial roughness

    :Returns:
        fresnel : Fresnel
            callable object for computing Fresnel reflectivity at Q

    Note that we do not correct for attenuation of the beam through the
    incident medium since we do not know the path length.
    """

    def __init__(self, rho=0, irho=0, sigma=0, Vrho=0, Virho=0):
        self.rho, self.Vrho, self.irho, self.Virho, self.sigma = rho, Vrho, irho, Virho, sigma

    def reflectivity(self, Q):
        """
        Compute the Fresnel reflectivity at the given Q/wavelength.
        """
        # Below we have the change in refractive index for entering through
        # the surface (delta_rho_Qp for Q positive), and through the substrate
        # (delta_rho_Qm for Q negative).  For Q negative we must negate the
        # change in scattering length density and ignore the absorption,
        # which should be handled by measuring the intensity through the
        # substrate, and therefore be corrected during reduction.
        delta_rho_Qp = 1e-6 * ((self.rho - self.Vrho) + 1j * self.irho)
        delta_rho_Qm = 1e-6 * ((self.Vrho - self.rho) + 1j * self.Virho)
        # print "fresnel", rho_Qp.shape, rho_Qm.shape, Q.shape

        delta_rho = choose(Q < 0, (delta_rho_Qp, delta_rho_Qm))
        kz = abs(Q) / 2
        f = sqrt(kz**2 - 4 * pi * delta_rho)  # fresnel coefficient

        # Compute reflectivity amplitude, with adjustment for roughness
        amp = (kz - f) / (kz + f) * exp(-2 * self.sigma**2 * kz * f)
        # Note: we do not need to check for a divide by zero.
        # Qc^2 = 16 pi rho.  Since rho is non-zero then Qc is non-zero.
        # For mu = 0:
        # * If |Qz| < Qc then f has an imaginary component, so |Qz|+f != 0.
        # * If |Qz| > Qc then |Qz| > 0 and f > 0, so |Qz|+f != 0.
        # * If |Qz| = Qc then |Qz| != 0 and f = 0, so |Qz|+f != 0.
        # For mu != 0:
        # * f has an imaginary component, so |Q|+f != 0.

        return (amp * conj(amp)).real

    # Make the reflectivity method the default
    __call__ = reflectivity


def test():
    import numpy as np
    from . import abeles

    # Rough silicon with an anomolously large absorbtion
    rho, irho = 2.07, 1.01
    Vrho, Virho = -1, 1.1
    sigma = 20
    fresnel = Fresnel(rho=rho, irho=irho, Vrho=Vrho, Virho=Virho, sigma=sigma)

    Mw = [0, 0]
    Mrho = [[Vrho, rho]]
    Mirho = [[Virho, irho]]
    Msigma = [sigma]

    Q = np.linspace(-0.1, 0.1, 101, "d")
    Rf = fresnel(Q)
    rm = abeles.refl(Q / 2, depth=Mw, rho=Mrho, irho=Mirho, sigma=Msigma)
    Rm = abs(rm) ** 2

    # print "Rm", Rm
    # print "Rf", Rf
    relerr = np.linalg.norm((Rf - Rm) / Rm)
    assert relerr < 1e-14, "relative error is %g" % relerr


if __name__ == "__main__":
    test()
